"""
Configuration of network device

.. versionadded:: 2016.11.0

:codeauthor:    Krzysztof Pawlowski <msciciel@msciciel.eu>
:maturity:      new
:depends:       python-ethtool
:platform:      linux

.. code-block:: yaml

    eth0:
      ethtool.coalesce:
        - name: eth0
        - rx_usecs: 24
        - tx_usecs: 48

    eth0:
      ethtool.ring:
        - name: eth0
        - rx: 1024
        - tx: 1024

    eth0:
      ethtool.offload:
        - name: eth0
        - tcp_segmentation_offload: on

"""


import logging

from salt.exceptions import CommandExecutionError

# Set up logging
log = logging.getLogger(__name__)


def __virtual__():
    """
    Provide ethtool state
    """
    if "ethtool.show_driver" in __salt__:
        return "ethtool"
    return (False, "ethtool module could not be loaded")


def coalesce(name, **kwargs):
    """
    Manage coalescing settings of network device

    name
        Interface name to apply coalescing settings

    .. code-block:: yaml

        eth0:
          ethtool.coalesce:
            - name: eth0
            - adaptive_rx: on
            - adaptive_tx: on
            - rx_usecs: 24
            - rx_frame: 0
            - rx_usecs_irq: 0
            - rx_frames_irq: 0
            - tx_usecs: 48
            - tx_frames: 0
            - tx_usecs_irq: 0
            - tx_frames_irq: 0
            - stats_block_usecs: 0
            - pkt_rate_low: 0
            - rx_usecs_low: 0
            - rx_frames_low: 0
            - tx_usecs_low: 0
            - tx_frames_low: 0
            - pkt_rate_high: 0
            - rx_usecs_high: 0
            - rx_frames_high: 0
            - tx_usecs_high: 0
            - tx_frames_high: 0
            - sample_interval: 0

    """
    ret = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": "Network device {} coalescing settings are up to date.".format(name),
    }
    apply_coalescing = False
    if "test" not in kwargs:
        kwargs["test"] = __opts__.get("test", False)

    # Build coalescing settings
    try:
        old = __salt__["ethtool.show_coalesce"](name)
        if not isinstance(old, dict):
            ret["result"] = False
            ret["comment"] = "Device {} coalescing settings are not supported".format(
                name
            )
            return ret

        new = {}
        diff = []

        # Retreive changes to made
        for key, value in kwargs.items():
            if key in old and value != old[key]:
                new.update({key: value})
                diff.append("{}: {}".format(key, value))

        # Dry run
        if kwargs["test"]:
            if not new:
                return ret
            if new:
                ret["result"] = None
                ret[
                    "comment"
                ] = "Device {} coalescing settings are set to be updated:\n{}".format(
                    name, "\n".join(diff)
                )
                return ret

        # Prepare return output
        if new:
            apply_coalescing = True
            ret["comment"] = "Device {} coalescing settings updated.".format(name)
            ret["changes"]["ethtool_coalesce"] = "\n".join(diff)

    except AttributeError as error:
        ret["result"] = False
        ret["comment"] = str(error)
        return ret

    # Apply coalescing settings
    if apply_coalescing:
        try:
            __salt__["ethtool.set_coalesce"](name, **new)
        except AttributeError as error:
            ret["result"] = False
            ret["comment"] = str(error)
            return ret

    return ret


def ring(name, **kwargs):
    """
    Manage rx/tx ring parameters of network device

    Use 'max' word to set with factory maximum

    name
        Interface name to apply ring parameters

    .. code-block:: yaml

        eth0:
          ethtool.ring:
            - name: eth0
            - rx: 1024
            - rx_mini: 0
            - rx_jumbo: 0
            - tx: max

    """
    ret = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": "Network device {} ring parameters are up to date.".format(name),
    }
    apply_ring = False
    if "test" not in kwargs:
        kwargs["test"] = __opts__.get("test", False)

    # Build ring parameters
    try:
        old = __salt__["ethtool.show_ring"](name)
        if not isinstance(old, dict):
            ret["result"] = False
            ret["comment"] = "Device {} ring parameters are not supported".format(name)
            return ret

        new = {}
        diff = []

        # Retreive changes to made
        for key, value in kwargs.items():
            if key in old:
                if value == "max":
                    value = old["{}_max".format(key)]

                if value != old[key]:
                    new.update({key: value})
                    diff.append("{}: {}".format(key, value))

        # Dry run
        if kwargs["test"]:
            if not new:
                return ret
            if new:
                ret["result"] = None
                ret[
                    "comment"
                ] = "Device {} ring parameters are set to be updated:\n{}".format(
                    name, "\n".join(diff)
                )
                return ret

        # Prepare return output
        if new:
            apply_ring = True
            ret["comment"] = "Device {} ring parameters updated.".format(name)
            ret["changes"]["ethtool_ring"] = "\n".join(diff)

    except AttributeError as error:
        ret["result"] = False
        ret["comment"] = str(error)
        return ret

    # Apply ring parameters
    if apply_ring:
        try:
            __salt__["ethtool.set_ring"](name, **new)
        except AttributeError as error:
            ret["result"] = False
            ret["comment"] = str(error)
            return ret

    return ret


def offload(name, **kwargs):
    """
    Manage protocol offload and other features of network device

    name
        Interface name to apply coalescing settings

    .. code-block:: yaml

        eth0:
          ethtool.offload:
            - name: eth0
            - tcp_segmentation_offload: on

    """
    ret = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": "Network device {} offload settings are up to date.".format(name),
    }
    apply_offload = False
    if "test" not in kwargs:
        kwargs["test"] = __opts__.get("test", False)

    # Build offload settings
    try:
        old = __salt__["ethtool.show_offload"](name)
        if not isinstance(old, dict):
            ret["result"] = False
            ret["comment"] = "Device {} offload settings are not supported".format(name)
            return ret

        new = {}
        diff = []

        # Retreive changes to made
        for key, value in kwargs.items():
            value = value and "on" or "off"
            if key in old and value != old[key]:
                new.update({key: value})
                diff.append("{}: {}".format(key, value))

        # Dry run
        if kwargs["test"]:
            if not new:
                return ret
            if new:
                ret["result"] = None
                ret[
                    "comment"
                ] = "Device {} offload settings are set to be updated:\n{}".format(
                    name, "\n".join(diff)
                )
                return ret

        # Prepare return output
        if new:
            apply_offload = True
            ret["comment"] = "Device {} offload settings updated.".format(name)
            ret["changes"]["ethtool_offload"] = "\n".join(diff)

    except AttributeError as error:
        ret["result"] = False
        ret["comment"] = str(error)
        return ret

    # Apply offload settings
    if apply_offload:
        try:
            __salt__["ethtool.set_offload"](name, **new)
        except AttributeError as error:
            ret["result"] = False
            ret["comment"] = str(error)
            return ret

    return ret


def pause(name, **kwargs):
    """
    .. versionadded:: 3006.0

    Manage pause parameters of network device

    name
        Interface name to apply pause parameters

    .. code-block:: yaml

        eth0:
          ethtool.pause:
            - name: eth0
            - autoneg: off
            - rx: off
            - tx: off

    """
    ret = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": "Network device {} pause parameters are up to date.".format(name),
    }
    apply_pause = False

    # Get current pause parameters
    try:
        old = __salt__["ethtool.show_pause"](name)
    except CommandExecutionError:
        ret["result"] = False
        ret["comment"] = "Device {} pause parameters are not supported".format(name)
        return ret

    # map ethtool command input to output text
    pause_map = {
        "autoneg": "Autonegotiate",
        "rx": "RX",
        "tx": "RX",
    }

    # Process changes
    new = {}
    diff = []

    for key, value in kwargs.items():
        key = key.lower()
        if key in pause_map:
            if value != old[pause_map[key]]:
                new.update({key: value})
                if value is True:
                    value = "on"
                elif value is False:
                    value = "off"
                diff.append("{}: {}".format(key, value))

    if not new:
        return ret

    # Dry run
    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = "Device {} pause parameters are set to be updated:\n{}".format(
            name, "\n".join(diff)
        )
        return ret

    # Apply pause parameters
    try:
        __salt__["ethtool.set_pause"](name, **new)
        # Prepare return output
        ret["comment"] = "Device {} pause parameters updated.".format(name)
        ret["changes"]["ethtool_pause"] = "\n".join(diff)
    except CommandExecutionError as exc:
        ret["result"] = False
        ret["comment"] = str(exc)
        return ret

    return ret



"""
    Test cases for salt.modules.status
"""


import os

import pytest

import salt.modules.status as status
import salt.utils.platform
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, Mock, mock_open, patch


@pytest.fixture
def configure_loader_modules():
    return {status: {}}


def test_boot_time_aix():
    mock = Mock()

    mock.return_value = {"stdout": "7-20:46:46", "retcode": 0}
    with patch.dict(status.__salt__, {"cmd.run_all": mock}):
        uptime = status._get_boot_time_aix()
        assert uptime == 679606

    mock.return_value = {"stdout": "20:46:46", "retcode": 0}
    with patch.dict(status.__salt__, {"cmd.run_all": mock}):
        uptime = status._get_boot_time_aix()
        assert uptime == 74806

    mock.return_value = {"stdout": "   01:32:37", "retcode": 0}
    with patch.dict(status.__salt__, {"cmd.run_all": mock}):
        uptime = status._get_boot_time_aix()
        assert uptime == 5557

    mock.return_value = {"stdout": "00:12", "retcode": 0}
    with patch.dict(status.__salt__, {"cmd.run_all": mock}):
        uptime = status._get_boot_time_aix()
        assert uptime == 12

    mock.return_value = {"stdout": "foo", "retcode": 0}
    with patch.dict(status.__salt__, {"cmd.run_all": mock}):
        with pytest.raises(CommandExecutionError):
            status._get_boot_time_aix()

    mock.return_value = {"stdout": "", "retcode": 1}
    with patch.dict(status.__salt__, {"cmd.run_all": mock}):
        with pytest.raises(CommandExecutionError):
            status._get_boot_time_aix()


def test_custom():
    mock = MagicMock()
    mock2 = MagicMock()

    mock.return_value = {
        "days": 2,
        "seconds": 249719,
        "since_iso": "2023-02-27T06:19:01.590002",
        "since_t": 1677478741,
        "time": "21:21",
        "users": 2,
    }
    # test pass correct info with correct config
    mock2.return_value = {"status.uptime.custom": ["days"]}
    with patch.dict(status.__salt__, {"config.dot_vals": mock2}):
        with patch.dict(status.__salt__, {"status.uptime": mock}):
            assert status.custom() == {"days": 2}

    # test pass correct info with incorrect config
    mock2.return_value = {"status.fail.custom": ["days"]}
    with patch.dict(status.__salt__, {"config.dot_vals": mock2}):
        with patch.dict(status.__salt__, {"status.uptime": mock}):
            assert status.custom() == {}

    # test incorrect info with correct config
    mock2.return_value = {"status.uptime.custom": ["day"]}
    with patch.dict(status.__salt__, {"config.dot_vals": mock2}):
        with patch.dict(status.__salt__, {"status.uptime": mock}):
            assert status.custom() == {"day": "UNKNOWN"}


def _set_up_test_uptime():
    """
    Define common mock data for status.uptime tests
    """

    class MockData:
        """
        Store mock data
        """

    m = MockData()
    m.now = 1477004312
    m.ut = 1540154.00
    m.idle = 3047777.32
    m.ret = {
        "users": 3,
        "seconds": 1540154,
        "since_t": 1475464158,
        "days": 17,
        "since_iso": "2016-10-03T03:09:18",
        "time": "19:49",
    }

    return m


def _set_up_test_uptime_sunos():
    """
    Define common mock data for cmd.run_all for status.uptime on SunOS
    """

    class MockData:
        """
        Store mock data
        """

    m = MockData()
    m.ret = {
        "retcode": 0,
        "stdout": "unix:0:system_misc:boot_time    1475464158",
    }

    return m


def test_uptime_linux():
    """
    Test modules.status.uptime function for Linux
    """
    m = _set_up_test_uptime()

    with patch.multiple(
        salt.utils.platform,
        is_linux=MagicMock(return_value=True),
        is_sunos=MagicMock(return_value=False),
        is_darwin=MagicMock(return_value=False),
        is_freebsd=MagicMock(return_value=False),
        is_openbsd=MagicMock(return_value=False),
        is_netbsd=MagicMock(return_value=False),
    ), patch("salt.utils.path.which", MagicMock(return_value=True)), patch.dict(
        status.__salt__,
        {"cmd.run": MagicMock(return_value=os.linesep.join(["1", "2", "3"]))},
    ), patch(
        "time.time", MagicMock(return_value=m.now)
    ), patch(
        "os.path.exists", MagicMock(return_value=True)
    ):
        proc_uptime = salt.utils.stringutils.to_str("{} {}".format(m.ut, m.idle))

        with patch("salt.utils.files.fopen", mock_open(read_data=proc_uptime)):
            ret = status.uptime()
            assert ret == m.ret
        with patch("os.path.exists", MagicMock(return_value=False)):
            with pytest.raises(CommandExecutionError):
                status.uptime()


def test_uptime_sunos():
    """
    Test modules.status.uptime function for SunOS
    """
    m = _set_up_test_uptime()
    m2 = _set_up_test_uptime_sunos()
    with patch.multiple(
        salt.utils.platform,
        is_linux=MagicMock(return_value=False),
        is_sunos=MagicMock(return_value=True),
        is_darwin=MagicMock(return_value=False),
        is_freebsd=MagicMock(return_value=False),
        is_openbsd=MagicMock(return_value=False),
        is_netbsd=MagicMock(return_value=False),
    ), patch("salt.utils.path.which", MagicMock(return_value=True)), patch.dict(
        status.__salt__,
        {
            "cmd.run": MagicMock(return_value=os.linesep.join(["1", "2", "3"])),
            "cmd.run_all": MagicMock(return_value=m2.ret),
        },
    ), patch(
        "time.time", MagicMock(return_value=m.now)
    ):
        ret = status.uptime()
        assert ret == m.ret


def test_uptime_macos():
    """
    Test modules.status.uptime function for macOS
    """
    m = _set_up_test_uptime()

    kern_boottime = (
        "{{ sec = {0}, usec = {1:0<6} }} Mon Oct 03 03:09:18.23 2016".format(
            *str(m.now - m.ut).split(".")
        )
    )
    with patch.multiple(
        salt.utils.platform,
        is_linux=MagicMock(return_value=False),
        is_sunos=MagicMock(return_value=False),
        is_darwin=MagicMock(return_value=True),
        is_freebsd=MagicMock(return_value=False),
        is_openbsd=MagicMock(return_value=False),
        is_netbsd=MagicMock(return_value=False),
    ), patch("salt.utils.path.which", MagicMock(return_value=True)), patch.dict(
        status.__salt__,
        {
            "cmd.run": MagicMock(return_value=os.linesep.join(["1", "2", "3"])),
            "sysctl.get": MagicMock(return_value=kern_boottime),
        },
    ), patch(
        "time.time", MagicMock(return_value=m.now)
    ):

        ret = status.uptime()
        assert ret == m.ret

        with patch.dict(status.__salt__, {"sysctl.get": MagicMock(return_value="")}):
            with pytest.raises(CommandExecutionError):
                status.uptime()


def test_uptime_return_success_not_supported():
    """
    Test modules.status.uptime function for other platforms
    """
    with patch.multiple(
        salt.utils.platform,
        is_linux=MagicMock(return_value=False),
        is_sunos=MagicMock(return_value=False),
        is_darwin=MagicMock(return_value=False),
        is_freebsd=MagicMock(return_value=False),
        is_openbsd=MagicMock(return_value=False),
        is_netbsd=MagicMock(return_value=False),
    ):
        exc_mock = MagicMock(side_effect=CommandExecutionError)
        with pytest.raises(CommandExecutionError):
            with patch.dict(status.__salt__, {"cmd.run": exc_mock}):
                status.uptime()


def _set_up_test_cpustats_openbsd():
    """
    Define mock data for status.cpustats on OpenBSD
    """

    class MockData:
        """
        Store mock data
        """

    m = MockData()
    m.ret = {
        "0": {
            "User": "0.0%",
            "Nice": "0.0%",
            "System": "4.5%",
            "Interrupt": "0.5%",
            "Idle": "95.0%",
        }
    }

    return m


def test_cpustats_openbsd():
    """
    Test modules.status.cpustats function for OpenBSD
    """
    m = _set_up_test_cpustats_openbsd()

    systat = (
        "\n\n   1 users Load 0.20 0.07 0.05                        salt.localdomain"
        " 09:42:42\nCPU                User           Nice        System    "
        " Interrupt          Idle\n0                  0.0%           0.0%         "
        " 4.5%          0.5%         95.0%\n"
    )

    with patch.multiple(
        salt.utils.platform,
        is_linux=MagicMock(return_value=False),
        is_sunos=MagicMock(return_value=False),
        is_darwin=MagicMock(return_value=False),
        is_freebsd=MagicMock(return_value=False),
        is_openbsd=MagicMock(return_value=True),
        is_netbsd=MagicMock(return_value=False),
    ), patch("salt.utils.path.which", MagicMock(return_value=True)), patch.dict(
        status.__grains__, {"kernel": "OpenBSD"}
    ), patch.dict(
        status.__salt__, {"cmd.run": MagicMock(return_value=systat)}
    ):
        ret = status.cpustats()
        assert ret == m.ret


def _set_up_test_cpuinfo_bsd():
    class MockData:
        """
        Store mock data
        """

    m = MockData()
    m.ret = {
        "hw.model": "Intel(R) Core(TM) i5-7287U CPU @ 3.30GHz",
        "hw.ncpu": "4",
    }

    return m


def test_cpuinfo_freebsd():
    m = _set_up_test_cpuinfo_bsd()
    sysctl = "hw.model:Intel(R) Core(TM) i5-7287U CPU @ 3.30GHz\nhw.ncpu:4"

    with patch.dict(status.__grains__, {"kernel": "FreeBSD"}):
        with patch.dict(status.__salt__, {"cmd.run": MagicMock(return_value=sysctl)}):
            ret = status.cpuinfo()
            assert ret == m.ret


def test_cpuinfo_openbsd():
    m = _set_up_test_cpuinfo_bsd()
    sysctl = "hw.model=Intel(R) Core(TM) i5-7287U CPU @ 3.30GHz\nhw.ncpu=4"

    for bsd in ["NetBSD", "OpenBSD"]:
        with patch.dict(status.__grains__, {"kernel": bsd}):
            with patch.dict(
                status.__salt__, {"cmd.run": MagicMock(return_value=sysctl)}
            ):
                ret = status.cpuinfo()
                assert ret == m.ret


def _set_up_test_meminfo_openbsd():
    class MockData:
        """
        Store mock data
        """

    m = MockData()
    m.ret = {
        "active virtual pages": "355M",
        "free list size": "305M",
        "page faults": "845",
        "pages reclaimed": "1",
        "pages paged in": "2",
        "pages paged out": "3",
        "pages freed": "4",
        "pages scanned": "5",
    }

    return m


def test_meminfo_openbsd():
    m = _set_up_test_meminfo_openbsd()
    vmstat = (
        " procs    memory       page                    disks    traps         "
        " cpu\n r   s   avm     fre  flt  re  pi  po  fr  sr cd0 sd0  int   sys  "
        " cs us sy id\n 2 103  355M    305M  845   1   2   3   4   5   0   1   21  "
        " 682   86  1  1 98"
    )

    with patch.dict(status.__grains__, {"kernel": "OpenBSD"}):
        with patch.dict(status.__salt__, {"cmd.run": MagicMock(return_value=vmstat)}):
            ret = status.meminfo()
            assert ret == m.ret


def _set_up_test_w_linux():
    """
    Define mock data for status.w on Linux
    """

    class MockData:
        """
        Store mock data
        """

    m = MockData()
    m.ret = [
        {
            "idle": "0s",
            "jcpu": "0.24s",
            "login": "13:42",
            "pcpu": "0.16s",
            "tty": "pts/1",
            "user": "root",
            "what": "nmap -sV 10.2.2.2",
        }
    ]

    return m


def _set_up_test_w_bsd():
    """
    Define mock data for status.w on Linux
    """

    class MockData:
        """
        Store mock data
        """

    m = MockData()
    m.ret = [
        {
            "idle": "0",
            "from": "10.2.2.1",
            "login": "1:42PM",
            "tty": "p1",
            "user": "root",
            "what": "nmap -sV 10.2.2.2",
        }
    ]

    return m


def test_w_linux():
    m = _set_up_test_w_linux()
    w_output = "root   pts/1  13:42    0s  0.24s  0.16s nmap -sV 10.2.2.2"

    with patch.dict(status.__grains__, {"kernel": "Linux"}):
        with patch.dict(status.__salt__, {"cmd.run": MagicMock(return_value=w_output)}):
            ret = status.w()
            assert ret == m.ret


def test_w_bsd():
    m = _set_up_test_w_bsd()
    w_output = "root   p1 10.2.2.1    1:42PM  0 nmap -sV 10.2.2.2"

    for bsd in ["Darwin", "FreeBSD", "OpenBSD"]:
        with patch.dict(status.__grains__, {"kernel": bsd}):
            with patch.dict(
                status.__salt__, {"cmd.run": MagicMock(return_value=w_output)}
            ):
                ret = status.w()
                assert ret == m.ret


def _set_up_test_status_pid_linux():
    class MockData:
        """
        Store mock data
        """

    m = MockData()
    m.ret = "2701\n7539\n7540\n7542\n7623"
    return m


def test_status_pid_linux():
    m = _set_up_test_status_pid_linux()
    ps = (
        "UID      PID PPID  C STIME TTY      TIME CMD\nroot     360    2  0 Jun08 ?"
        "    00:00:00   [jbd2/dm-0-8]\nroot     947    2  0 Jun08 ?    00:00:00  "
        " [jbd2/dm-1-8]\nroot     949    2  0 Jun08 ?    00:00:09  "
        " [jbd2/dm-3-8]\nroot     951    2  0 Jun08 ?    00:00:00  "
        " [jbd2/dm-4-8]\nroot    2701    1  0 Jun08 ?    00:00:28   /usr/sbin/httpd"
        " -k start\napache  7539 2701  0 04:40 ?    00:00:04     /usr/sbin/httpd -k"
        " start\napache  7540 2701  0 04:40 ?    00:00:02     /usr/sbin/httpd -k"
        " start\napache  7542 2701  0 04:40 ?    00:01:46     /usr/sbin/httpd -k"
        " start\napache  7623 2701  0 04:40 ?    00:02:41     /usr/sbin/httpd -k"
        " start\nroot    1564    1  0 Jun11 ?    00:07:19   /usr/bin/python3"
        " /usr/bin/salt-minion -d\nroot    6674 1564  0 19:53 ?    00:00:00    "
        " /usr/bin/python3 /usr/bin/salt-call status.pid httpd -l debug"
    )

    with patch.dict(status.__grains__, {"ps": "ps -efHww"}):
        with patch.dict(
            status.__salt__, {"cmd.run_stdout": MagicMock(return_value=ps)}
        ):
            with patch.object(os, "getpid", return_value="6674"):
                ret = status.pid("httpd")
                assert ret == m.ret



# pylint: disable=resource-leakage
"""
This is a shim that handles checking and updating salt thin and
then invoking thin.

This is not intended to be instantiated as a module, rather it is a
helper script used by salt.client.ssh.Single.  It is here, in a
separate file, for convenience of development.
"""
from __future__ import absolute_import, print_function

import hashlib
import os
import shutil
import stat
import subprocess
import sys
import tarfile
import time

THIN_ARCHIVE = "salt-thin.tgz"
EXT_ARCHIVE = "salt-ext_mods.tgz"

# Keep these in sync with salt/defaults/exitcodes.py
EX_THIN_PYTHON_INVALID = 10
EX_THIN_DEPLOY = 11
EX_THIN_CHECKSUM = 12
EX_MOD_DEPLOY = 13
EX_SCP_NOT_FOUND = 14
EX_CANTCREAT = 73


class OptionsContainer:
    """
    An empty class for holding instance attribute values.
    """


OPTIONS = OptionsContainer()
ARGS = None
# The below line is where OPTIONS can be redefined with internal options
# (rather than cli arguments) when the shim is bundled by
# client.ssh.Single._cmd_str()
# pylint: disable=block-comment-should-start-with-cardinal-space
#%%OPTS


def get_system_encoding():
    """
    Get system encoding. Most of this code is a part of salt/__init__.py
    """
    # This is the most trustworthy source of the system encoding, though, if
    # salt is being imported after being daemonized, this information is lost
    # and reset to None
    encoding = None

    if not sys.platform.startswith("win") and sys.stdin is not None:
        # On linux we can rely on sys.stdin for the encoding since it
        # most commonly matches the filesystem encoding. This however
        # does not apply to windows
        encoding = sys.stdin.encoding

    if not encoding:
        # If the system is properly configured this should return a valid
        # encoding. MS Windows has problems with this and reports the wrong
        # encoding
        import locale

        try:
            encoding = locale.getencoding()
        except AttributeError:
            # Python < 3.11
            encoding = locale.getpreferredencoding(do_setlocale=True)

        # This is now garbage collectable
        del locale

        if not encoding:
            # This is most likely ascii which is not the best but we were
            # unable to find a better encoding. If this fails, we fall all
            # the way back to ascii
            encoding = sys.getdefaultencoding()
        if not encoding:
            if sys.platform.startswith("darwin"):
                # Mac OS X uses UTF-8
                encoding = "utf-8"
            elif sys.platform.startswith("win"):
                # Windows uses a configurable encoding; on Windows, Python uses the name "mbcs"
                # to refer to whatever the currently configured encoding is.
                encoding = "mbcs"
            else:
                # On linux default to ascii as a last resort
                encoding = "ascii"
    return encoding


def is_windows():
    """
    Simple function to return if a host is Windows or not
    """
    return sys.platform.startswith("win")


def need_deployment():
    """
    Salt thin needs to be deployed - prep the target directory and emit the
    delimiter and exit code that signals a required deployment.
    """
    if os.path.exists(OPTIONS.saltdir):
        shutil.rmtree(OPTIONS.saltdir)
    old_umask = os.umask(0o077)  # pylint: disable=blacklisted-function
    try:
        os.makedirs(OPTIONS.saltdir)
    finally:
        os.umask(old_umask)  # pylint: disable=blacklisted-function
    # Verify perms on saltdir
    if not is_windows():
        euid = os.geteuid()
        dstat = os.stat(OPTIONS.saltdir)
        if dstat.st_uid != euid:
            # Attack detected, try again
            need_deployment()
        # AIX has non-POSIX bit 0o240700, isolate to 0o40700
        if dstat.st_mode & ~65536 != 16832:
            # Attack detected
            need_deployment()
        # If SUDOing then also give the super user group write permissions
        sudo_gid = os.environ.get("SUDO_GID")
        if sudo_gid:
            try:
                os.chown(OPTIONS.saltdir, -1, int(sudo_gid))
                stt = os.stat(OPTIONS.saltdir)
                os.chmod(
                    OPTIONS.saltdir,
                    stt.st_mode | stat.S_IWGRP | stat.S_IRGRP | stat.S_IXGRP,
                )
            except OSError:
                sys.stdout.write(
                    "\n\nUnable to set permissions on thin directory.\nIf sudo_user is"
                    " set and is not root, be certain the user is in the same group\nas"
                    " the login user"
                )
                sys.exit(1)

    # Delimiter emitted on stdout *only* to indicate shim message to master.
    sys.stdout.write("{0}\ndeploy\n".format(OPTIONS.delimiter))
    sys.exit(EX_THIN_DEPLOY)


# Adapted from salt.utils.hashutils.get_hash()
def get_hash(path, form="sha1", chunk_size=4096):
    """
    Generate a hash digest string for a file.
    """
    try:
        hash_type = getattr(hashlib, form)
    except AttributeError:
        raise ValueError("Invalid hash type: {0}".format(form))
    with open(path, "rb") as ifile:
        hash_obj = hash_type()
        # read the file in in chunks, not the entire file
        for chunk in iter(lambda: ifile.read(chunk_size), b""):
            hash_obj.update(chunk)
        return hash_obj.hexdigest()


def unpack_thin(thin_path):
    """
    Unpack the Salt thin archive.
    """
    tfile = tarfile.TarFile.gzopen(thin_path)
    old_umask = os.umask(0o077)  # pylint: disable=blacklisted-function
    tfile.extractall(path=OPTIONS.saltdir)
    tfile.close()
    os.umask(old_umask)  # pylint: disable=blacklisted-function
    try:
        os.unlink(thin_path)
    except OSError:
        pass
    reset_time(OPTIONS.saltdir)


def need_ext():
    """
    Signal that external modules need to be deployed.
    """
    sys.stdout.write("{0}\next_mods\n".format(OPTIONS.delimiter))
    sys.exit(EX_MOD_DEPLOY)


def unpack_ext(ext_path):
    """
    Unpack the external modules.
    """
    modcache = os.path.join(
        OPTIONS.saltdir, "running_data", "var", "cache", "salt", "minion", "extmods"
    )
    tfile = tarfile.TarFile.gzopen(ext_path)
    old_umask = os.umask(0o077)  # pylint: disable=blacklisted-function
    tfile.extractall(path=modcache)
    tfile.close()
    os.umask(old_umask)  # pylint: disable=blacklisted-function
    os.unlink(ext_path)
    ver_path = os.path.join(modcache, "ext_version")
    ver_dst = os.path.join(OPTIONS.saltdir, "ext_version")
    shutil.move(ver_path, ver_dst)


def reset_time(path=".", amt=None):
    """
    Reset atime/mtime on all files to prevent systemd swipes only part of the files in the /tmp.
    """
    if not amt:
        amt = int(time.time())
    for fname in os.listdir(path):
        fname = os.path.join(path, fname)
        if os.path.isdir(fname):
            reset_time(fname, amt=amt)
        os.utime(
            fname,
            (
                amt,
                amt,
            ),
        )


def get_executable():
    """
    Find executable which matches supported python version in the thin
    """
    pymap = {}
    with open(os.path.join(OPTIONS.saltdir, "supported-versions")) as _fp:
        for line in _fp.readlines():
            ns, v_maj, v_min = line.strip().split(":")
            pymap[ns] = (int(v_maj), int(v_min))

    pycmds = (
        sys.executable,
        "python3",
        "python27",
        "python2.7",
        "python26",
        "python2.6",
        "python2",
        "python",
    )
    for py_cmd in pycmds:
        stdout, _ = subprocess.Popen(
            [
                py_cmd,
                "-c",
                "import sys; sys.stdout.write('%s:%s' % (sys.version_info[0], sys.version_info[1]))",
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        ).communicate()
        if sys.version_info[0] == 2 and sys.version_info[1] < 7:
            stdout = stdout.decode(get_system_encoding(), "replace").strip()
        else:
            stdout = stdout.decode(
                encoding=get_system_encoding(), errors="replace"
            ).strip()
        if not stdout:
            continue
        c_vn = tuple([int(x) for x in stdout.split(":")])
        for ns in pymap:
            if (
                c_vn[0] == pymap[ns][0]
                and c_vn >= pymap[ns]
                and os.path.exists(os.path.join(OPTIONS.saltdir, ns))
            ):
                return py_cmd

    sys.exit(EX_THIN_PYTHON_INVALID)


def main(argv):  # pylint: disable=W0613
    """
    Main program body
    """
    thin_path = os.path.join(OPTIONS.saltdir, THIN_ARCHIVE)
    if os.path.isfile(thin_path):
        if OPTIONS.checksum != get_hash(thin_path, OPTIONS.hashfunc):
            need_deployment()
        unpack_thin(thin_path)
        # Salt thin now is available to use
    else:
        if not sys.platform.startswith("win"):
            scpstat = subprocess.Popen(["/bin/sh", "-c", "command -v scp"]).wait()
            if scpstat != 0:
                sys.exit(EX_SCP_NOT_FOUND)

        if os.path.exists(OPTIONS.saltdir) and not os.path.isdir(OPTIONS.saltdir):
            sys.stderr.write(
                'ERROR: salt path "{0}" exists but is not a directory\n'.format(
                    OPTIONS.saltdir
                )
            )
            sys.exit(EX_CANTCREAT)

        if not os.path.exists(OPTIONS.saltdir):
            need_deployment()

        code_checksum_path = os.path.normpath(
            os.path.join(OPTIONS.saltdir, "code-checksum")
        )
        if not os.path.exists(code_checksum_path) or not os.path.isfile(
            code_checksum_path
        ):
            sys.stderr.write(
                "WARNING: Unable to locate current code checksum: {0}.\n".format(
                    code_checksum_path
                )
            )
            need_deployment()
        with open(code_checksum_path, "r") as vpo:
            cur_code_cs = vpo.readline().strip()
        if cur_code_cs != OPTIONS.code_checksum:
            sys.stderr.write(
                "WARNING: current code checksum {0} is different to {1}.\n".format(
                    cur_code_cs, OPTIONS.code_checksum
                )
            )
            need_deployment()
        # Salt thin exists and is up-to-date - fall through and use it

    salt_call_path = os.path.join(OPTIONS.saltdir, "salt-call")
    if not os.path.isfile(salt_call_path):
        sys.stderr.write('ERROR: thin is missing "{0}"\n'.format(salt_call_path))
        need_deployment()

    with open(os.path.join(OPTIONS.saltdir, "minion"), "w") as config:
        config.write(OPTIONS.config + "\n")
    if OPTIONS.ext_mods:
        ext_path = os.path.join(OPTIONS.saltdir, EXT_ARCHIVE)
        if os.path.exists(ext_path):
            unpack_ext(ext_path)
        else:
            version_path = os.path.join(OPTIONS.saltdir, "ext_version")
            if not os.path.exists(version_path) or not os.path.isfile(version_path):
                need_ext()
            with open(version_path, "r") as vpo:
                cur_version = vpo.readline().strip()
            if cur_version != OPTIONS.ext_mods:
                need_ext()
    # Fix parameter passing issue
    if len(ARGS) == 1:
        argv_prepared = ARGS[0].split()
    else:
        argv_prepared = ARGS

    salt_argv = [
        get_executable(),
        salt_call_path,
        "--retcode-passthrough",
        "--local",
        "--metadata",
        "--out",
        "json",
        "-l",
        "quiet",
        "-c",
        OPTIONS.saltdir,
    ]

    try:
        if argv_prepared[-1].startswith("--no-parse="):
            salt_argv.append(argv_prepared.pop(-1))
    except (IndexError, TypeError):
        pass

    salt_argv.append("--")
    salt_argv.extend(argv_prepared)

    sys.stderr.write("SALT_ARGV: {0}\n".format(salt_argv))

    # Only emit the delimiter on *both* stdout and stderr when completely successful.
    # Yes, the flush() is necessary.
    sys.stdout.write(OPTIONS.delimiter + "\n")
    sys.stdout.flush()
    if not OPTIONS.tty:
        sys.stderr.write(OPTIONS.delimiter + "\n")
        sys.stderr.flush()
    if OPTIONS.cmd_umask is not None:
        old_umask = os.umask(OPTIONS.cmd_umask)  # pylint: disable=blacklisted-function
    if OPTIONS.tty:
        proc = subprocess.Popen(
            salt_argv, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        # Returns bytes instead of string on python 3
        stdout, _ = proc.communicate()
        sys.stdout.write(
            stdout.decode(encoding=get_system_encoding(), errors="replace")
        )
        sys.stdout.flush()
        retcode = proc.returncode
        if OPTIONS.wipe:
            shutil.rmtree(OPTIONS.saltdir)
    elif OPTIONS.wipe:
        retcode = subprocess.call(salt_argv)
        shutil.rmtree(OPTIONS.saltdir)
    else:
        retcode = subprocess.call(salt_argv)
    if OPTIONS.cmd_umask is not None:
        os.umask(old_umask)  # pylint: disable=blacklisted-function
    return retcode


if __name__ == "__main__":
    sys.exit(main(sys.argv))



import pathlib

import pytest

import salt.spm.pkgfiles.local as spm
import salt.syspaths
from tests.support.mock import MagicMock


@pytest.fixture()
def configure_loader_modules():
    return {spm: {"__opts__": {"spm_node_type": "master"}}}


class MockTar:
    def __init__(self):
        self.name = str(pathlib.Path("apache", "_README"))
        self.path = str(pathlib.Path(salt.syspaths.CACHE_DIR, "master", "extmods"))


def test_install_file(tmp_path):
    """
    test spm.pkgfiles.local
    """
    assert (
        spm.install_file(
            "apache",
            formula_tar=MagicMock(),
            member=MockTar(),
            formula_def={"name": "apache"},
            conn={"formula_path": str(tmp_path / "test")},
        )
        == MockTar().path
    )



import pytest

import salt.modules.nilrt_ip as nilrt_ip
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, mock, patch
from tests.support.unit import TestCase

pytest.importorskip("pyiface")


class NilrtIPTestCase(TestCase, LoaderModuleMockMixin):
    """
    TestCase for salt.modules.nilrt_ip module
    """

    def setup_loader_modules(self):
        return {nilrt_ip: {"__grains__": {"lsb_distrib_id": "not_nilrt"}}}

    def test_change_state_down_state(self):
        """
        Tests _change_state when not connected
        and new state is down
        """
        iface_mock = MagicMock()
        iface_mock.name = "test_interface"
        with patch("pyiface.getIfaces", return_value=[iface_mock]):
            with patch(
                "salt.modules.nilrt_ip._change_dhcp_config", return_value=True
            ) as change_dhcp_config_mock:
                assert nilrt_ip._change_state("test_interface", "down")
                change_dhcp_config_mock.assert_called_with("test_interface", False)

    def test_change_state_up_state(self):
        """
        Tests _change_state when connected
        and new state is up
        """
        iface_mock = MagicMock()
        iface_mock.name = "test_interface"
        with patch("pyiface.getIfaces", return_value=[iface_mock]):
            with patch(
                "salt.modules.nilrt_ip._change_dhcp_config", return_value=True
            ) as change_dhcp_config_mock:
                assert nilrt_ip._change_state("test_interface", "up")
                change_dhcp_config_mock.assert_called_with("test_interface")

    def test_set_static_all_with_dns(self):
        """
        Tests set_static_all with dns provided
        """
        save_config_mock = MagicMock(return_value=None)
        with patch.dict(nilrt_ip.__grains__, {"lsb_distrib_id": "nilrt"}):
            with patch.object(
                nilrt_ip, "_get_adapter_mode_info", MagicMock(return_value="TCPIP")
            ):
                with patch.object(nilrt_ip, "_save_config", save_config_mock):
                    with patch.object(
                        nilrt_ip, "_restart", MagicMock(return_value=None)
                    ):
                        self.assertEqual(
                            True,
                            nilrt_ip.set_static_all(
                                "eth0",
                                "192.168.10.4",
                                "255.255.255.0",
                                "192.168.10.1",
                                "8.8.4.4 8.8.8.8",
                            ),
                        )
                        mock_call = mock.call("eth0", "DNS_Address", "8.8.4.4")
                        self.assertEqual(True, mock_call in save_config_mock.mock_calls)

    def test_set_static_all_no_dns(self):
        """
        Tests set_static_all with no dns provided
        """
        save_config_mock = MagicMock(return_value=None)
        with patch.dict(nilrt_ip.__grains__, {"lsb_distrib_id": "nilrt"}):
            with patch.object(
                nilrt_ip, "_get_adapter_mode_info", MagicMock(return_value="TCPIP")
            ):
                with patch.object(nilrt_ip, "_save_config", save_config_mock):
                    with patch.object(
                        nilrt_ip, "_restart", MagicMock(return_value=None)
                    ):
                        self.assertEqual(
                            True,
                            nilrt_ip.set_static_all(
                                "eth0", "192.168.10.4", "255.255.255.0", "192.168.10.1"
                            ),
                        )
                        self.assertEqual(
                            None,
                            next(
                                (
                                    call
                                    for call in save_config_mock.mock_calls
                                    if "DNS_Address" in call.args
                                ),
                                None,
                            ),
                        )



import time
import uuid

import attr
from pytestshellutils.utils import ports
from saltfactories.daemons.container import SaltMinion

from tests.conftest import CODE_DIR


@attr.s(kw_only=True, slots=True)
class SaltVirtMinionContainerFactory(SaltMinion):

    host_uuid = attr.ib(default=attr.Factory(uuid.uuid4))
    ssh_port = attr.ib(
        default=attr.Factory(ports.get_unused_localhost_port), repr=False
    )
    sshd_port = attr.ib(default=attr.Factory(ports.get_unused_localhost_port))
    libvirt_tcp_port = attr.ib(
        default=attr.Factory(ports.get_unused_localhost_port), repr=False
    )
    libvirt_tls_port = attr.ib(
        default=attr.Factory(ports.get_unused_localhost_port), repr=False
    )

    uri = attr.ib(init=False)
    ssh_uri = attr.ib(init=False)
    tcp_uri = attr.ib(init=False)
    tls_uri = attr.ib(init=False)

    def __attrs_post_init__(self):
        self.uri = "localhost:{}".format(self.sshd_port)
        self.ssh_uri = "qemu+ssh://{}/system".format(self.uri)
        self.tcp_uri = "qemu+tcp://localhost:{}/system".format(self.libvirt_tcp_port)
        self.tls_uri = "qemu+tls://127.0.0.1:{}/system".format(self.libvirt_tls_port)

        if "environment" not in self.container_run_kwargs:
            self.container_run_kwargs["environment"] = {}
        self.container_run_kwargs["environment"].update(
            {
                "SSH_PORT": str(self.ssh_port),
                "SSHD_PORT": str(self.sshd_port),
                "LIBVIRT_TCP_PORT": str(self.libvirt_tcp_port),
                "LIBVIRT_TLS_PORT": str(self.libvirt_tls_port),
                "NO_START_MINION": "1",
                "HOST_UUID": self.host_uuid,
                "PYTHONDONTWRITEBYTECODE": "1",
                "PYTHONPATH": str(CODE_DIR),
            }
        )
        super().__attrs_post_init__()
        if "volumes" not in self.container_run_kwargs:
            self.container_run_kwargs["volumes"] = {}
        self.container_run_kwargs["volumes"].update(
            {
                str(CODE_DIR): {"bind": "/salt", "mode": "z"},
                str(CODE_DIR): {"bind": str(CODE_DIR), "mode": "z"},
            }
        )
        self.container_run_kwargs["working_dir"] = str(CODE_DIR)
        self.container_run_kwargs["network_mode"] = "host"
        self.container_run_kwargs["cap_add"] = ["ALL"]
        self.container_run_kwargs["privileged"] = True
        self.python_executable = "python3"
        self.container_start_check(self._check_script_path_exists)
        for port in (self.sshd_port, self.libvirt_tcp_port, self.libvirt_tls_port):
            self.check_ports[port] = port

    def _check_script_path_exists(self, timeout_at):
        while time.time() <= timeout_at:
            # Once we're able to ls the salt-minion script it means the container
            # has salt installed
            ret = self.run("ls", "-lah", self.get_script_path())
            if ret.returncode == 0:
                break
            time.sleep(1)
        else:
            return False
        return True



"""
    Test cases for salt.states.kernelpkg
"""


import pytest

import salt.states.kernelpkg as kernelpkg
from tests.support.mock import MagicMock, patch


@pytest.fixture
def kernel_list():
    return ["4.4.0-70-generic", "4.4.0-71-generic", "4.5.1-14-generic"]


@pytest.fixture
def state_name():
    return "kernelpkg-test"


@pytest.fixture
def configure_loader_modules():
    return {
        kernelpkg: {
            "__salt__": {
                "system.reboot": MagicMock(return_value=None),
                "kernelpkg.upgrade": MagicMock(
                    return_value={
                        "upgrades": {"kernel": {"old": "1.0.0", "new": "2.0.0"}}
                    }
                ),
                "kernelpkg.active": MagicMock(return_value=0),
                "kernelpkg.latest_installed": MagicMock(return_value=0),
            }
        }
    }


def test_latest_installed_with_changes(kernel_list, state_name):
    """
    Test - latest_installed when an upgrade is available
    """
    installed = MagicMock(return_value=kernel_list[:-1])
    upgrade = MagicMock(return_value=kernel_list[-1])
    with patch.dict(kernelpkg.__salt__, {"kernelpkg.list_installed": installed}):
        with patch.dict(kernelpkg.__salt__, {"kernelpkg.latest_available": upgrade}):
            with patch.dict(kernelpkg.__opts__, {"test": False}):
                kernelpkg.__salt__["kernelpkg.upgrade"].reset_mock()
                ret = kernelpkg.latest_installed(name=state_name)
                assert ret["name"] == state_name
                assert ret["result"]
                assert isinstance(ret["changes"], dict)
                assert isinstance(ret["comment"], str)
                kernelpkg.__salt__["kernelpkg.upgrade"].assert_called_once()

            with patch.dict(kernelpkg.__opts__, {"test": True}):
                kernelpkg.__salt__["kernelpkg.upgrade"].reset_mock()
                ret = kernelpkg.latest_installed(name=state_name)
                assert ret["name"] == state_name
                assert ret["result"] is None
                assert ret["changes"] == {}
                assert isinstance(ret["comment"], str)
                kernelpkg.__salt__["kernelpkg.upgrade"].assert_not_called()


def test_latest_installed_at_latest(kernel_list, state_name):
    """
    Test - latest_installed when no upgrade is available
    """
    installed = MagicMock(return_value=kernel_list)
    upgrade = MagicMock(return_value=kernel_list[-1])
    with patch.dict(kernelpkg.__salt__, {"kernelpkg.list_installed": installed}):
        with patch.dict(kernelpkg.__salt__, {"kernelpkg.latest_available": upgrade}):
            with patch.dict(kernelpkg.__opts__, {"test": False}):
                ret = kernelpkg.latest_installed(name=state_name)
                assert ret["name"] == state_name
                assert ret["result"]
                assert ret["changes"] == {}
                assert isinstance(ret["comment"], str)
                kernelpkg.__salt__["kernelpkg.upgrade"].assert_not_called()

            with patch.dict(kernelpkg.__opts__, {"test": True}):
                ret = kernelpkg.latest_installed(name=state_name)
                assert ret["name"] == state_name
                assert ret["result"]
                assert ret["changes"] == {}
                assert isinstance(ret["comment"], str)
                kernelpkg.__salt__["kernelpkg.upgrade"].assert_not_called()


def test_latest_active_with_changes(state_name):
    """
    Test - latest_active when a new kernel is available
    """
    reboot = MagicMock(return_value=True)
    latest = MagicMock(return_value=1)
    with patch.dict(
        kernelpkg.__salt__,
        {"kernelpkg.needs_reboot": reboot, "kernelpkg.latest_installed": latest},
    ), patch.dict(kernelpkg.__opts__, {"test": False}):
        kernelpkg.__salt__["system.reboot"].reset_mock()
        ret = kernelpkg.latest_active(name=state_name)
        assert ret["name"] == state_name
        assert ret["result"]
        assert isinstance(ret["changes"], dict)
        assert isinstance(ret["comment"], str)
        kernelpkg.__salt__["system.reboot"].assert_called_once()

        with patch.dict(kernelpkg.__opts__, {"test": True}):
            kernelpkg.__salt__["system.reboot"].reset_mock()
            ret = kernelpkg.latest_active(name=state_name)
            assert ret["name"] == state_name
            assert ret["result"] is None
            assert ret["changes"] == {"kernel": {"new": 1, "old": 0}}
            assert isinstance(ret["comment"], str)
            kernelpkg.__salt__["system.reboot"].assert_not_called()


def test_latest_active_at_latest(state_name):
    """
    Test - latest_active when the newest kernel is already active
    """
    reboot = MagicMock(return_value=False)
    with patch.dict(kernelpkg.__salt__, {"kernelpkg.needs_reboot": reboot}):
        with patch.dict(kernelpkg.__opts__, {"test": False}):
            kernelpkg.__salt__["system.reboot"].reset_mock()
            ret = kernelpkg.latest_active(name=state_name)
            assert ret["name"] == state_name
            assert ret["result"]
            assert ret["changes"] == {}
            assert isinstance(ret["comment"], str)
            kernelpkg.__salt__["system.reboot"].assert_not_called()

        with patch.dict(kernelpkg.__opts__, {"test": True}):
            kernelpkg.__salt__["system.reboot"].reset_mock()
            ret = kernelpkg.latest_active(name=state_name)
            assert ret["name"] == state_name
            assert ret["result"]
            assert ret["changes"] == {}
            assert isinstance(ret["comment"], str)
            kernelpkg.__salt__["system.reboot"].assert_not_called()


def test_latest_wait(state_name):
    """
    Test - latest_wait static results
    """
    ret = kernelpkg.latest_wait(name=state_name)
    assert ret["name"] == state_name
    assert ret["result"]
    assert ret["changes"] == {}
    assert isinstance(ret["comment"], str)



# flake8: noqa
# pylint: skip-file
from __future__ import absolute_import, division, print_function
from salt.ext.tornado.test.util import unittest


class ImportTest(unittest.TestCase):
    def test_import_everything(self):
        # Some of our modules are not otherwise tested.  Import them
        # all (unless they have external dependencies) here to at
        # least ensure that there are no syntax errors.
        import salt.ext.tornado.auth
        import salt.ext.tornado.autoreload
        import salt.ext.tornado.concurrent
        import salt.ext.tornado.escape
        import salt.ext.tornado.gen
        import salt.ext.tornado.http1connection
        import salt.ext.tornado.httpclient
        import salt.ext.tornado.httpserver
        import salt.ext.tornado.httputil
        import salt.ext.tornado.ioloop
        import salt.ext.tornado.iostream
        import salt.ext.tornado.locale
        import salt.ext.tornado.log
        import salt.ext.tornado.netutil
        import salt.ext.tornado.options
        import salt.ext.tornado.process
        import salt.ext.tornado.simple_httpclient
        import salt.ext.tornado.stack_context
        import salt.ext.tornado.tcpserver
        import salt.ext.tornado.tcpclient
        import salt.ext.tornado.template
        import salt.ext.tornado.testing
        import salt.ext.tornado.util
        import salt.ext.tornado.web
        import salt.ext.tornado.websocket
        import salt.ext.tornado.wsgi

    # for modules with dependencies, if those dependencies can be loaded,
    # load them too.

    def test_import_pycurl(self):
        try:
            import pycurl  # type: ignore
        except ImportError:
            pass
        else:
            import salt.ext.tornado.curl_httpclient



"""
    tests.pytests.unit.beacons.test_glxinfo
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    glxinfo beacon test cases
"""
import pytest

import salt.beacons.glxinfo as glxinfo
from tests.support.mock import Mock, patch


@pytest.fixture
def configure_loader_modules():
    return {glxinfo: {"last_state": {}}}


def test_no_glxinfo_command():
    with patch("salt.utils.path.which") as mock:
        mock.return_value = None

        ret = glxinfo.__virtual__()

        mock.assert_called_once_with("glxinfo")
        assert ret == (False, "glxinfo is missing.")


def test_with_glxinfo_command():
    with patch("salt.utils.path.which") as mock:
        mock.return_value = "/usr/bin/glxinfo"

        ret = glxinfo.__virtual__()

        mock.assert_called_once_with("glxinfo")
        assert ret == "glxinfo"


def test_non_list_config():
    config = {}

    ret = glxinfo.validate(config)
    assert ret == (False, "Configuration for glxinfo beacon must be a list.")


def test_no_user():
    config = [{"screen_event": True}]

    _expected = (
        False,
        "Configuration for glxinfo beacon must include a user as glxinfo is not"
        " available to root.",
    )

    ret = glxinfo.validate(config)
    assert ret == _expected


def test_screen_state():
    config = [{"screen_event": True, "user": "frank"}]

    mock = Mock(return_value=0)
    with patch.dict(glxinfo.__salt__, {"cmd.retcode": mock}):
        ret = glxinfo.validate(config)
        assert ret == (True, "Valid beacon configuration")

        ret = glxinfo.beacon(config)
        assert ret == [{"tag": "screen_event", "screen_available": True}]
        mock.assert_called_once_with(
            "DISPLAY=:0 glxinfo", runas="frank", python_shell=True
        )


def test_screen_state_missing():
    config = [{"screen_event": True, "user": "frank"}]

    mock = Mock(return_value=255)
    with patch.dict(glxinfo.__salt__, {"cmd.retcode": mock}):
        ret = glxinfo.validate(config)
        assert ret == (True, "Valid beacon configuration")

        ret = glxinfo.beacon(config)
        assert ret == [{"tag": "screen_event", "screen_available": False}]


def test_screen_state_no_repeat():
    config = [{"screen_event": True, "user": "frank"}]

    mock = Mock(return_value=255)
    with patch.dict(glxinfo.__salt__, {"cmd.retcode": mock}):
        ret = glxinfo.validate(config)
        assert ret == (True, "Valid beacon configuration")

        ret = glxinfo.beacon(config)
        assert ret == [{"tag": "screen_event", "screen_available": False}]

        ret = glxinfo.beacon(config)
        assert ret == []


def test_screen_state_change():
    config = [{"screen_event": True, "user": "frank"}]

    mock = Mock(side_effect=[255, 0])
    with patch.dict(glxinfo.__salt__, {"cmd.retcode": mock}):
        ret = glxinfo.validate(config)
        assert ret == (True, "Valid beacon configuration")

        ret = glxinfo.beacon(config)
        assert ret == [{"tag": "screen_event", "screen_available": False}]

        ret = glxinfo.beacon(config)
        assert ret == [{"tag": "screen_event", "screen_available": True}]



"""
Module used to access the esxdatacenter proxy connection methods

.. Warning::
    This module will be deprecated in a future release of Salt. VMware strongly
    recommends using the
    `VMware Salt extensions <https://docs.saltproject.io/salt/extensions/salt-ext-modules-vmware/en/latest/all.html>`_
    instead of the ESX data center module. Because the Salt extensions are newer
    and actively supported by VMware, they are more compatible with current
    versions of ESXi and they work well with the latest features in the VMware
    product line.

"""

import logging
from functools import wraps

import salt.utils.platform

log = logging.getLogger(__name__)

__proxyenabled__ = ["esxdatacenter"]
# Define the module's virtual name
__virtualname__ = "esxdatacenter"


def __virtual__():
    """
    Only work on proxy
    """
    if salt.utils.platform.is_proxy():
        return __virtualname__
    return (False, "Must be run on a proxy minion")


def _deprecation_message(function):
    """
    Decorator wrapper to warn about azurearm deprecation
    """

    @wraps(function)
    def wrapped(*args, **kwargs):
        salt.utils.versions.warn_until(
            "Argon",
            "The 'esxdatacenter' functionality in Salt has been deprecated and its "
            "functionality will be removed in version 3008 in favor of the "
            "saltext.vmware Salt Extension. "
            "(https://github.com/saltstack/salt-ext-modules-vmware)",
            category=FutureWarning,
        )
        ret = function(*args, **salt.utils.args.clean_kwargs(**kwargs))
        return ret

    return wrapped


@_deprecation_message
def get_details():
    return __proxy__["esxdatacenter.get_details"]()



"""
Tests for orchestration events
"""
import concurrent.futures
import functools
import json
import logging
import time

import attr
import pytest
from saltfactories.utils import random_string

import salt.utils.jid
import salt.utils.platform
import salt.utils.pycrypto

log = logging.getLogger(__name__)


@attr.s(kw_only=True, slots=True)
class TestMasterAccount:
    username = attr.ib()
    password = attr.ib()
    _delete_account = attr.ib(init=False, repr=False, default=False)

    @username.default
    def _default_username(self):
        return random_string("account-", uppercase=False)

    @password.default
    def _default_password(self):
        return random_string("pwd-", size=8)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass


@pytest.fixture(scope="session")
def salt_auth_account_m_factory():
    return TestMasterAccount(username="saltdev-auth-m")


@pytest.fixture(scope="module")
def salt_auth_account_m(salt_auth_account_m_factory):
    with salt_auth_account_m_factory as account:
        yield account


@pytest.fixture(scope="module")
def runner_master_config(salt_auth_account_m):
    return {
        "external_auth": {
            "pam": {salt_auth_account_m.username: [{"*": [".*"]}, "@runner", "@wheel"]}
        }
    }


@pytest.fixture(scope="module")
def runner_salt_master(salt_factories, runner_master_config):
    factory = salt_factories.salt_master_daemon(
        "runner-master", defaults=runner_master_config
    )
    with factory.started():
        yield factory


@pytest.fixture(scope="module")
def runner_salt_run_cli(runner_salt_master):
    return runner_salt_master.salt_run_cli()


@pytest.fixture(scope="module")
def runner_salt_call_cli(runner_salt_minion):
    return runner_salt_minion.salt_call_cli()


@pytest.fixture(scope="module")
def runner_add_user(runner_salt_run_cli, salt_auth_account_m):
    ## create user on master to use
    ret = runner_salt_run_cli.run("salt.cmd", "user.add", salt_auth_account_m.username)
    assert ret.returncode == 0

    yield

    ## remove user on master
    ret = runner_salt_run_cli.run(
        "salt.cmd", "user.delete", salt_auth_account_m.username
    )
    assert ret.returncode == 0


def test_state_event(salt_run_cli, salt_cli, salt_minion):
    """
    test to ensure state.event
    runner returns correct data
    """
    # If and when we have tornado 5.0, we can just do io_loop.run_in_executor() ....
    with concurrent.futures.ThreadPoolExecutor(max_workers=3) as executor:
        runner_future = executor.submit(
            functools.partial(
                salt_run_cli.run, "state.event", "salt/job/*/new", count=1
            )
        )

        while not runner_future.done():
            ret = salt_cli.run("--static", "test.ping", minion_tgt=salt_minion.id)
            assert ret.returncode == 0
            assert ret.data is True

        # Wait for the runner command which should now have data to return to us
        exc = runner_future.exception()
        if exc:
            raise exc
        ret = runner_future.result()
        assert ret.returncode == 0
        # We have to parse the JSON ourselves since we have regular output mixed with JSON output
        data = None
        for line in ret.stdout.splitlines():
            try:
                _, payload = line.strip().split("\t")
                data = json.loads(payload)
                break
            except ValueError:
                continue
        assert isinstance(data, dict)
        assert salt_minion.id in data["minions"]


def test_jid_in_ret_event(salt_run_cli, salt_master, salt_minion, event_listener):
    """
    Test to confirm that the ret event for the orchestration contains the
    jid for the jobs spawned.
    """
    test_state_contents = """
    date:
      cmd.run
    """
    test_orch_contents = """
    date_cmd:
      salt.state:
        - tgt: {minion_id}
        - sls: test-state

    ping_minion:
      salt.function:
        - name: test.ping
        - tgt: {minion_id}

    fileserver.file_list:
      salt.runner

    config.values:
      salt.wheel
    """.format(
        minion_id=salt_minion.id
    )
    with salt_master.state_tree.base.temp_file(
        "test-state.sls", test_state_contents
    ), salt_master.state_tree.base.temp_file("test-orch.sls", test_orch_contents):
        start_time = time.time()
        jid = salt.utils.jid.gen_jid(salt_master.config)

        ret = salt_run_cli.run("--jid", jid, "state.orchestrate", "test-orch")
        assert ret.returncode == 0
        orch_job_data = ret.data
        for step_data in orch_job_data["data"][salt_master.id].values():
            assert "__jid__" in step_data

        expected_event_tag = "salt/run/{}/ret".format(jid)
        event_pattern = (salt_master.id, expected_event_tag)

        matched_events = event_listener.wait_for_events(
            [event_pattern], after_time=start_time, timeout=120
        )
        assert (
            matched_events.found_all_events
        ), "Failed to receive the event with the tag '{}'".format(expected_event_tag)
        for event in matched_events.matches:
            for job_data in event.data["return"]["data"][salt_master.id].values():
                assert "__jid__" in job_data


# This test is flaky on FreeBSD
@pytest.mark.skip_on_freebsd
@pytest.mark.skip_on_spawning_platform(
    reason="The '__low__' global is not populated on spawning platforms"
)
def test_parallel_orchestrations(
    salt_run_cli, salt_master, salt_minion, event_listener
):
    """
    Test to confirm that the parallel state requisite works in orch
    we do this by running 19 test.sleep's of 10 seconds, and a last 10 seconds sleep
    which depends on the previous 19.
    It should take more than 20 seconds and less than 19*10 seconds
    """
    test_orch_contents = """
    {% for count in range(1, 20) %}

    sleep {{ count }}:
        module.run:
            - name: test.sleep
            - length: 10
            - parallel: True

    {% endfor %}

    sleep 21:
        module.run:
            - name: test.sleep
            - length: 10
            - parallel: True
            - require:
                - module: sleep 1
    """
    with salt_master.state_tree.base.temp_file("test-orch.sls", test_orch_contents):
        start_time = time.time()
        jid = salt.utils.jid.gen_jid(salt_master.config)

        ret = salt_run_cli.run("--jid", jid, "state.orchestrate", "test-orch")
        assert ret.returncode == 0
        orch_job_data = ret.data
        for step_data in orch_job_data["data"][salt_master.id].values():
            # we expect each duration to be greater than 10s
            assert step_data["duration"] > 10 * 1000

        # Since we started range(1, 20)(19) sleep state steps, and the last state
        # step requires all of those to have finished before it runs, since we are
        # running in parallel, the duration should be more than 20 seconds, and
        # less than 19*10(190) seconds, less then half actually
        duration = time.time() - start_time
        assert duration > 20
        assert duration < 19 * 10 / 2

        expected_event_tag = "salt/run/{}/ret".format(jid)
        event_pattern = (salt_master.id, expected_event_tag)

        matched_events = event_listener.wait_for_events(
            [event_pattern], after_time=start_time, timeout=120
        )
        assert (
            matched_events.found_all_events
        ), "Failed to receive the event with the tag '{}'".format(expected_event_tag)
        for event in matched_events.matches:
            for job_data in event.data["return"]["data"][salt_master.id].values():
                # we expect each duration to be greater than 10s
                assert job_data["duration"] > 10 * 1000

        # Since we started range(1, 20)(19) sleep state steps, and the last state
        # step requires all of those to have finished before it runs, since we are
        # running in parallel, the duration should be more than 20 seconds, and
        # less than 19*10(190) seconds, less then half actually
        duration = time.time() - start_time
        assert duration > 20
        assert duration < 19 * 10 / 2


def test_orchestration_soft_kill(salt_run_cli, salt_master):
    sls_contents = """
    stage_one:
        test.succeed_without_changes

    stage_two:
        test.fail_without_changes
    """
    with salt_master.state_tree.base.temp_file("test-orch.sls", sls_contents):
        jid = salt.utils.jid.gen_jid(salt_master.config)

        # Without soft kill, the orchestration will fail because stage_two is set to fail
        ret = salt_run_cli.run("--jid", jid, "state.orchestrate", "test-orch")
        assert ret.returncode == 1
        for state_data in ret.data["data"][salt_master.id].values():
            if state_data["__id__"] == "stage_two":
                assert state_data["result"] is False
            else:
                assert state_data["result"] is True

        # With soft kill set, 'stage_two' will not run, thus, the orchestration will not fail
        # and 'stage_two' will not be present in the returned data
        jid = salt.utils.jid.gen_jid(salt_master.config)
        ret = salt_run_cli.run("state.soft_kill", jid, "stage_two")
        assert ret.returncode == 0
        ret = salt_run_cli.run("--jid", jid, "state.orchestrate", "test-orch")
        assert ret.returncode == 0
        for state_data in ret.data["data"][salt_master.id].values():
            if state_data["__id__"] == "stage_two":
                pytest.fail("'stage_two' was present in the ochestration return data")
            else:
                assert state_data["result"] is True


def test_orchestration_with_pillar_dot_items(salt_run_cli, salt_master):
    """
    Test to confirm when using a state file that includes other state file, if
    one of those state files includes pillar related functions that will not
    be pulling from the pillar cache that all the state files are available and
    the file_roots has been preserved.  See issues #48277 and #46986.
    """
    main_sls_contents = """
    include:
      - one
      - two
      - three
    """
    one_sls_contents = """
    {%- set foo = salt['saltutil.runner']('pillar.show_pillar') %}
    placeholder_one:
      test.succeed_without_changes
    """
    two_sls_contents = """
    placeholder_two:
      test.succeed_without_changes
    """
    three_sls_contents = """
    placeholder_three:
      test.succeed_without_changes
    """
    with salt_master.state_tree.base.temp_file(
        "test-orch.sls", main_sls_contents
    ), salt_master.state_tree.base.temp_file(
        "one.sls", one_sls_contents
    ), salt_master.state_tree.base.temp_file(
        "two.sls", two_sls_contents
    ), salt_master.state_tree.base.temp_file(
        "three.sls", three_sls_contents
    ):
        jid = salt.utils.jid.gen_jid(salt_master.config)

        ret = salt_run_cli.run("--jid", jid, "state.orchestrate", "test-orch")
        assert ret.returncode == 0
        for state_data in ret.data["data"][salt_master.id].values():
            # Each state should be successful
            assert state_data["result"] is True


def test_orchestration_onchanges_and_prereq(
    salt_run_cli, salt_master, salt_minion, tmp_path
):
    sls_contents = """
    manage_a_file:
      salt.state:
        - tgt: {minion_id}
        - sls:
          - orch-req-test

    do_onchanges:
      salt.function:
        - tgt: {minion_id}
        - name: test.ping
        - onchanges:
          - salt: manage_a_file

    do_prereq:
      salt.function:
        - tgt: {minion_id}
        - name: test.ping
        - prereq:
          - salt: manage_a_file
    """.format(
        minion_id=salt_minion.id
    )

    orch_test_file = tmp_path / "orch-test-file"
    req_sls_contents = """
    {}:
      file.managed:
        - contents: 'Hello World!'
    """.format(
        orch_test_file
    )

    with salt_master.state_tree.base.temp_file(
        "test-orch.sls", sls_contents
    ), salt_master.state_tree.base.temp_file("orch-req-test.sls", req_sls_contents):
        jid1 = salt.utils.jid.gen_jid(salt_master.config)

        # Run in test mode, will describe what changes would occur
        ret = salt_run_cli.run(
            "--jid", jid1, "state.orchestrate", "test-orch", test=True
        )
        assert ret.returncode == 0
        ret1 = ret.data

        # Now run without test mode to actually create the file
        ret = salt_run_cli.run("state.orchestrate", "test-orch")
        assert ret.returncode == 0

        # Run again in test mode. Since there were no changes, the requisites should not fire.
        jid2 = salt.utils.jid.gen_jid(salt_master.config)
        ret = salt_run_cli.run(
            "--jid", jid2, "state.orchestrate", "test-orch", test=True
        )
        assert ret.returncode == 0
        ret2 = ret.data

        # The first time through, all three states should have a None result
        for state_data in ret1["data"][salt_master.id].values():
            assert state_data["result"] is None
            if state_data.get("__id__") == "manage_a_file":
                # The file.managed state should have shown changes in the test mode
                # return data.
                assert state_data["changes"]

        # While the second time through, they should all have a True result.
        for state_data in ret2["data"][salt_master.id].values():
            assert state_data["result"] is True
            if state_data.get("__id__") == "manage_a_file":
                # After the file was created, running again in test mode should have
                # shown no changes.
                assert not state_data["changes"]


@pytest.mark.slow_test
@pytest.mark.skip_if_not_root
@pytest.mark.skip_on_windows
@pytest.mark.skip_on_darwin
def test_unknown_in_runner_event(
    runner_salt_run_cli,
    runner_salt_master,
    salt_minion,
    salt_auth_account_m,
    runner_add_user,
    event_listener,
):
    """
    Test to confirm that the ret event for the orchestration contains the
    jid for the jobs spawned.
    """
    file_roots_base_dir = runner_salt_master.config["file_roots"]["base"][0]
    test_top_file_contents = """
    base:
      '{minion_id}':
        - {file_roots}
    """.format(
        minion_id=salt_minion.id, file_roots=file_roots_base_dir
    )
    test_init_state_contents = """
    always-passes-with-any-kwarg:
      test.nop:
        - name: foo
        - something: else
        - foo: bar
    always-passes:
      test.succeed_without_changes:
        - name: foo
    always-changes-and-succeeds:
      test.succeed_with_changes:
        - name: foo
    {{slspath}}:
      test.nop
    """
    test_orch_contents = """
    test_highstate:
      salt.state:
        - tgt: {minion_id}
        - highstate: True
    test_runner_metasyntetic:
      salt.runner:
        - name: test.metasyntactic
        - locality: us
    """.format(
        minion_id=salt_minion.id
    )
    with runner_salt_master.state_tree.base.temp_file(
        "top.sls", test_top_file_contents
    ), runner_salt_master.state_tree.base.temp_file(
        "init.sls", test_init_state_contents
    ), runner_salt_master.state_tree.base.temp_file(
        "orch.sls", test_orch_contents
    ):
        ret = runner_salt_run_cli.run(
            "salt.cmd", "shadow.gen_password", salt_auth_account_m.password
        )
        assert ret.returncode == 0

        gen_pwd = ret.stdout
        ret = runner_salt_run_cli.run(
            "salt.cmd", "shadow.set_password", salt_auth_account_m.username, gen_pwd
        )
        assert ret.returncode == 0

        jid = salt.utils.jid.gen_jid(runner_salt_master.config)
        start_time = time.time()

        ret = runner_salt_run_cli.run(
            "--jid",
            jid,
            "-a",
            "pam",
            "--username",
            salt_auth_account_m.username,
            "--password",
            salt_auth_account_m.password,
            "state.orchestrate",
            "orch",
        )
        assert not ret.stdout.startswith("Authentication failure")

        expected_new_event_tag = "salt/run/*/new"
        event_pattern = (runner_salt_master.id, expected_new_event_tag)
        found_events = event_listener.get_events([event_pattern], after_time=start_time)

        for event in found_events:
            if event.data["fun"] == "runner.test.metasyntactic":
                assert event.data["user"] == salt_auth_account_m.username

        expected_ret_event_tag = "salt/run/*/ret"
        event_pattern = (runner_salt_master.id, expected_ret_event_tag)
        found_events = event_listener.get_events([event_pattern], after_time=start_time)

        for event in found_events:
            if event.data["fun"] == "runner.test.metasyntactic":
                assert event.data["user"] == salt_auth_account_m.username



"""
    Test cases for salt.modules.win_powercfg
"""


import pytest

import salt.modules.win_powercfg as powercfg
from tests.support.mock import MagicMock, call, patch


@pytest.fixture
def configure_loader_modules():
    return {powercfg: {"__grains__": {"osrelease": 8}}}


@pytest.fixture
def query_output():
    return """Subgroup GUID: 238c9fa8-0aad-41ed-83f4-97be242c8f20  (Hibernate)
            GUID Alias: SUB_SLEEP
            Power Setting GUID: 29f6c1db-86da-48c5-9fdb-f2b67b1f44da  (Hibernate after)
            GUID Alias: HIBERNATEIDLE
            Minimum Possible Setting: 0x00000000
            Maximum Possible Setting: 0xffffffff
            Possible Settings increment: 0x00000001
            Possible Settings units: Seconds
            Current AC Power Setting Index: 0x00000708
            Current DC Power Setting Index: 0x00000384"""


def test_set_monitor_timeout(query_output):
    """
    Test to make sure we can set the monitor timeout value
    """
    mock = MagicMock(return_value=0)
    mock.side_effect = [
        "Power Scheme GUID: 381b4222-f694-41f0-9685-ff5bb260df2e  (Balanced)",
        query_output,
    ]

    mock_retcode = MagicMock(return_value=0)

    with patch.dict(powercfg.__salt__, {"cmd.run": mock}):
        with patch.dict(powercfg.__salt__, {"cmd.retcode": mock_retcode}):
            powercfg.set_monitor_timeout(0, "dc")
            mock.assert_called_once_with(
                "powercfg /getactivescheme", python_shell=False
            )
            mock_retcode.assert_called_once_with(
                "powercfg /setdcvalueindex 381b4222-f694-41f0-9685-ff5bb260df2e"
                " SUB_VIDEO VIDEOIDLE 0",
                python_shell=False,
            )


def test_set_disk_timeout(query_output):
    """
    Test to make sure we can set the disk timeout value
    """
    mock = MagicMock()
    mock.side_effect = [
        "Power Scheme GUID: 381b4222-f694-41f0-9685-ff5bb260df2e  (Balanced)",
        query_output,
    ]

    mock_retcode = MagicMock(return_value=0)

    with patch.dict(powercfg.__salt__, {"cmd.run": mock}):
        with patch.dict(powercfg.__salt__, {"cmd.retcode": mock_retcode}):
            powercfg.set_disk_timeout(0, "dc")
            mock.assert_called_once_with(
                "powercfg /getactivescheme", python_shell=False
            )
            mock_retcode.assert_called_once_with(
                "powercfg /setdcvalueindex 381b4222-f694-41f0-9685-ff5bb260df2e"
                " SUB_DISK DISKIDLE 0",
                python_shell=False,
            )


def test_set_standby_timeout(query_output):
    """
    Test to make sure we can set the standby timeout value
    """
    mock = MagicMock(return_value=0)
    mock.side_effect = [
        "Power Scheme GUID: 381b4222-f694-41f0-9685-ff5bb260df2e  (Balanced)",
        query_output,
    ]

    mock_retcode = MagicMock(return_value=0)

    with patch.dict(powercfg.__salt__, {"cmd.run": mock}):
        with patch.dict(powercfg.__salt__, {"cmd.retcode": mock_retcode}):
            powercfg.set_standby_timeout(0, "dc")
            mock.assert_called_once_with(
                "powercfg /getactivescheme", python_shell=False
            )
            mock_retcode.assert_called_once_with(
                "powercfg /setdcvalueindex 381b4222-f694-41f0-9685-ff5bb260df2e"
                " SUB_SLEEP STANDBYIDLE 0",
                python_shell=False,
            )


def test_set_hibernate_timeout(query_output):
    """
    Test to make sure we can set the hibernate timeout value
    """
    mock = MagicMock(return_value=0)
    mock.side_effect = [
        "Power Scheme GUID: 381b4222-f694-41f0-9685-ff5bb260df2e  (Balanced)",
        query_output,
    ]

    mock_retcode = MagicMock(return_value=0)

    with patch.dict(powercfg.__salt__, {"cmd.run": mock}):
        with patch.dict(powercfg.__salt__, {"cmd.retcode": mock_retcode}):
            powercfg.set_hibernate_timeout(0, "dc")
            mock.assert_called_once_with(
                "powercfg /getactivescheme", python_shell=False
            )
            mock_retcode.assert_called_once_with(
                "powercfg /setdcvalueindex 381b4222-f694-41f0-9685-ff5bb260df2e"
                " SUB_SLEEP HIBERNATEIDLE 0",
                python_shell=False,
            )


def test_get_monitor_timeout(query_output):
    """
    Test to make sure we can get the monitor timeout value
    """
    mock = MagicMock()
    mock.side_effect = [
        "Power Scheme GUID: 381b4222-f694-41f0-9685-ff5bb260df2e  (Balanced)",
        query_output,
    ]

    with patch.dict(powercfg.__salt__, {"cmd.run": mock}):
        ret = powercfg.get_monitor_timeout()
        calls = [
            call("powercfg /getactivescheme", python_shell=False),
            call(
                "powercfg /q 381b4222-f694-41f0-9685-ff5bb260df2e SUB_VIDEO"
                " VIDEOIDLE",
                python_shell=False,
            ),
        ]
        mock.assert_has_calls(calls)

        assert {"ac": 30, "dc": 15} == ret


def test_get_disk_timeout(query_output):
    """
    Test to make sure we can get the disk timeout value
    """
    mock = MagicMock()
    mock.side_effect = [
        "Power Scheme GUID: 381b4222-f694-41f0-9685-ff5bb260df2e  (Balanced)",
        query_output,
    ]

    with patch.dict(powercfg.__salt__, {"cmd.run": mock}):
        ret = powercfg.get_disk_timeout()
        calls = [
            call("powercfg /getactivescheme", python_shell=False),
            call(
                "powercfg /q 381b4222-f694-41f0-9685-ff5bb260df2e SUB_DISK" " DISKIDLE",
                python_shell=False,
            ),
        ]
        mock.assert_has_calls(calls)

        assert {"ac": 30, "dc": 15} == ret


def test_get_standby_timeout(query_output):
    """
    Test to make sure we can get the standby timeout value
    """
    mock = MagicMock()
    mock.side_effect = [
        "Power Scheme GUID: 381b4222-f694-41f0-9685-ff5bb260df2e  (Balanced)",
        query_output,
    ]

    with patch.dict(powercfg.__salt__, {"cmd.run": mock}):
        ret = powercfg.get_standby_timeout()
        calls = [
            call("powercfg /getactivescheme", python_shell=False),
            call(
                "powercfg /q 381b4222-f694-41f0-9685-ff5bb260df2e SUB_SLEEP"
                " STANDBYIDLE",
                python_shell=False,
            ),
        ]
        mock.assert_has_calls(calls)

        assert {"ac": 30, "dc": 15} == ret


def test_get_hibernate_timeout(query_output):
    """
    Test to make sure we can get the hibernate timeout value
    """
    mock = MagicMock()
    mock.side_effect = [
        "Power Scheme GUID: 381b4222-f694-41f0-9685-ff5bb260df2e  (Balanced)",
        query_output,
    ]

    with patch.dict(powercfg.__salt__, {"cmd.run": mock}):
        ret = powercfg.get_hibernate_timeout()
        calls = [
            call("powercfg /getactivescheme", python_shell=False),
            call(
                "powercfg /q 381b4222-f694-41f0-9685-ff5bb260df2e SUB_SLEEP"
                " HIBERNATEIDLE",
                python_shell=False,
            ),
        ]
        mock.assert_has_calls(calls)

        assert {"ac": 30, "dc": 15} == ret


def test_windows_7(query_output):
    """
    Test to make sure we can get the hibernate timeout value on windows 7
    """
    mock = MagicMock()
    mock.side_effect = [
        "Power Scheme GUID: 381b4222-f694-41f0-9685-ff5bb260df2e  (Balanced)",
        query_output,
    ]

    with patch.dict(powercfg.__salt__, {"cmd.run": mock}):
        with patch.dict(powercfg.__grains__, {"osrelease": "7"}):
            ret = powercfg.get_hibernate_timeout()
            calls = [
                call("powercfg /getactivescheme", python_shell=False),
                call(
                    "powercfg /q 381b4222-f694-41f0-9685-ff5bb260df2e SUB_SLEEP",
                    python_shell=False,
                ),
            ]
            mock.assert_has_calls(calls)

            assert {"ac": 30, "dc": 15} == ret


def test_set_hibernate_timeout_scheme(query_output):
    """
    Test to make sure we can set the hibernate timeout value
    """
    mock = MagicMock(return_value=0)
    mock.side_effect = [query_output]

    with patch.dict(powercfg.__salt__, {"cmd.retcode": mock}):
        powercfg.set_hibernate_timeout(0, "dc", scheme="SCHEME_MIN")
        mock.assert_called_once_with(
            "powercfg /setdcvalueindex SCHEME_MIN SUB_SLEEP HIBERNATEIDLE 0",
            python_shell=False,
        )


def test_get_hibernate_timeout_scheme(query_output):
    """
    Test to make sure we can get the hibernate timeout value with a
    specified scheme
    """
    mock = MagicMock()
    mock.side_effect = [query_output]

    with patch.dict(powercfg.__salt__, {"cmd.run": mock}):
        ret = powercfg.get_hibernate_timeout(scheme="SCHEME_MIN")
        mock.assert_called_once_with(
            "powercfg /q SCHEME_MIN SUB_SLEEP HIBERNATEIDLE", python_shell=False
        )

        assert {"ac": 30, "dc": 15} == ret



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>

    Test cases for salt.modules.nftables
"""
import json

import pytest

import salt.modules.nftables as nftables
import salt.utils.files
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, mock_open, patch


@pytest.fixture
def configure_loader_modules():
    return {nftables: {}}


# 'version' function tests: 1


def test_version():
    """
    Test if it return version from nftables --version
    """
    mock = MagicMock(return_value="nf_tables 0.3-1")
    with patch.dict(nftables.__salt__, {"cmd.run": mock}):
        assert nftables.version() == "0.3-1"


# 'build_rule' function tests: 1


def test_build_rule():
    """
    Test if it build a well-formatted nftables rule based on kwargs.
    """
    assert nftables.build_rule(full="True") == {
        "result": False,
        "rule": "",
        "comment": "Table needs to be specified",
    }

    assert nftables.build_rule(table="filter", full="True") == {
        "result": False,
        "rule": "",
        "comment": "Chain needs to be specified",
    }

    assert nftables.build_rule(table="filter", chain="input", full="True") == {
        "result": False,
        "rule": "",
        "comment": "Command needs to be specified",
    }

    assert nftables.build_rule(
        table="filter",
        chain="input",
        command="insert",
        position="3",
        full="True",
    ) == {
        "result": True,
        "rule": "nft insert rule ip filter input position 3 ",
        "comment": "Successfully built rule",
    }

    assert nftables.build_rule(
        table="filter", chain="input", command="insert", full="True"
    ) == {
        "result": True,
        "rule": "nft insert rule ip filter input ",
        "comment": "Successfully built rule",
    }

    assert nftables.build_rule(
        table="filter", chain="input", command="halt", full="True"
    ) == {
        "result": True,
        "rule": "nft halt rule ip filter input ",
        "comment": "Successfully built rule",
    }

    assert nftables.build_rule(
        table="filter",
        chain="input",
        command="insert",
        position="3",
        full="True",
        connstate="related,established",
        saddr="10.0.0.1",
        daddr="10.0.0.2",
        jump="accept",
    ) == {
        "result": True,
        "rule": (
            "nft insert rule ip filter input position 3 ct state {"
            " related,established } ip saddr 10.0.0.1 ip daddr 10.0.0.2 accept"
        ),
        "comment": "Successfully built rule",
    }

    assert nftables.build_rule() == {"result": True, "rule": "", "comment": ""}


# 'get_saved_rules' function tests: 1


def test_get_saved_rules():
    """
    Test if it return a data structure of the rules in the conf file
    """
    with patch.dict(nftables.__grains__, {"os_family": "Debian"}):
        with patch.object(salt.utils.files, "fopen", MagicMock(mock_open())):
            assert nftables.get_saved_rules() == []


# 'list_tables' function tests: 1


def test_list_tables():
    """
    Test if it return a data structure of the current, in-memory tables
    """
    list_tables = [{"family": "inet", "name": "filter", "handle": 2}]
    list_tables_mock = MagicMock(return_value=list_tables)

    with patch.object(nftables, "list_tables", list_tables_mock):
        assert nftables.list_tables() == list_tables

    list_tables_mock = MagicMock(return_value=[])
    with patch.object(nftables, "list_tables", list_tables_mock):
        assert nftables.list_tables() == []


# 'get_rules' function tests: 1


def test_get_rules():
    """
    Test if it return a data structure of the current, in-memory rules
    """
    list_tables_mock = MagicMock(
        return_value=[{"family": "inet", "name": "filter", "handle": 2}]
    )
    list_rules_return = """table inet filter {
        chain input {
            type filter hook input priority 0; policy accept;
        }

        chain forward {
            type filter hook forward priority 0; policy accept;
        }

        chain output {
            type filter hook output priority 0; policy accept;
        }
    }"""
    list_rules_mock = MagicMock(return_value=list_rules_return)
    expected = [list_rules_return]

    with patch.object(nftables, "list_tables", list_tables_mock):
        with patch.dict(nftables.__salt__, {"cmd.run": list_rules_mock}):
            assert nftables.get_rules() == expected

    list_tables_mock = MagicMock(return_value=[])
    with patch.object(nftables, "list_tables", list_tables_mock):
        assert nftables.get_rules() == []


# 'get_rules_json' function tests: 1


def test_get_rules_json():
    """
    Test if it return a data structure of the current, in-memory rules
    """
    list_rules_return = """
    {
      "nftables": [
        {
          "table": {
            "family": "ip",
            "name": "filter",
            "handle": 47
          }
        },
        {
          "chain": {
            "family": "ip",
            "table": "filter",
            "name": "input",
            "handle": 1,
            "type": "filter",
            "hook": "input",
            "prio": 0,
            "policy": "accept"
          }
        },
        {
          "chain": {
            "family": "ip",
            "table": "filter",
            "name": "forward",
            "handle": 2,
            "type": "filter",
            "hook": "forward",
            "prio": 0,
            "policy": "accept"
          }
        },
        {
          "chain": {
            "family": "ip",
            "table": "filter",
            "name": "output",
            "handle": 3,
            "type": "filter",
            "hook": "output",
            "prio": 0,
            "policy": "accept"
          }
        }
      ]
    }
    """
    list_rules_mock = MagicMock(return_value=list_rules_return)
    expected = json.loads(list_rules_return)["nftables"]

    with patch.dict(nftables.__salt__, {"cmd.run": list_rules_mock}):
        assert nftables.get_rules_json() == expected

    list_rules_mock = MagicMock(return_value=[])
    with patch.dict(nftables.__salt__, {"cmd.run": list_rules_mock}):
        assert nftables.get_rules_json() == []


# 'save' function tests: 1


def test_save():
    """
    Test if it save the current in-memory rules to disk
    """
    with patch.dict(nftables.__grains__, {"os_family": "Debian"}):
        mock = MagicMock(return_value=False)
        with patch.dict(nftables.__salt__, {"file.directory_exists": mock}):
            with patch.dict(nftables.__salt__, {"cmd.run": mock}):
                with patch.object(salt.utils.files, "fopen", MagicMock(mock_open())):
                    assert nftables.save() == "#! nft -f\n\n"

                with patch.object(
                    salt.utils.files, "fopen", MagicMock(side_effect=IOError)
                ):
                    pytest.raises(CommandExecutionError, nftables.save)


# 'get_rule_handle' function tests: 1


def test_get_rule_handle():
    """
    Test if it get the handle for a particular rule
    """
    assert nftables.get_rule_handle() == {
        "result": False,
        "comment": "Chain needs to be specified",
    }

    assert nftables.get_rule_handle(chain="input") == {
        "result": False,
        "comment": "Rule needs to be specified",
    }

    _ru = "input tcp dport 22 log accept"
    ret = {"result": False, "comment": "Table filter in family ipv4 does not exist"}
    mock = MagicMock(return_value="")
    with patch.dict(nftables.__salt__, {"cmd.run": mock}):
        assert nftables.get_rule_handle(chain="input", rule=_ru) == ret

    ret = {
        "result": False,
        "comment": "Chain input in table filter in family ipv4 does not exist",
    }
    mock = MagicMock(return_value="table ip filter")
    with patch.dict(nftables.__salt__, {"cmd.run": mock}):
        assert nftables.get_rule_handle(chain="input", rule=_ru) == ret

    ret = {
        "result": False,
        "comment": (
            "Rule input tcp dport 22 log accept chain"
            " input in table filter in family ipv4 does not exist"
        ),
    }
    ret1 = {
        "result": False,
        "comment": "Could not find rule input tcp dport 22 log accept",
    }
    with patch.object(
        nftables,
        "check_table",
        MagicMock(return_value={"result": True, "comment": ""}),
    ):
        with patch.object(
            nftables,
            "check_chain",
            MagicMock(return_value={"result": True, "comment": ""}),
        ):
            _ret1 = {
                "result": False,
                "comment": (
                    "Rule input tcp dport 22 log accept"
                    " chain input in table filter in"
                    " family ipv4 does not exist"
                ),
            }
            _ret2 = {"result": True, "comment": ""}
            with patch.object(nftables, "check", MagicMock(side_effect=[_ret1, _ret2])):
                assert nftables.get_rule_handle(chain="input", rule=_ru) == ret

                _ru = "input tcp dport 22 log accept"
                mock = MagicMock(return_value="")
                with patch.dict(nftables.__salt__, {"cmd.run": mock}):
                    assert nftables.get_rule_handle(chain="input", rule=_ru) == ret1


# 'check' function tests: 1


def test_check():
    """
    Test if it check for the existence of a rule in the table and chain
    """
    assert nftables.check() == {
        "result": False,
        "comment": "Chain needs to be specified",
    }

    assert nftables.check(chain="input") == {
        "result": False,
        "comment": "Rule needs to be specified",
    }

    _ru = "tcp dport 22 log accept"
    ret = {"result": False, "comment": "Table filter in family ipv4 does not exist"}
    mock = MagicMock(return_value="")
    with patch.dict(nftables.__salt__, {"cmd.run": mock}):
        assert nftables.check(chain="input", rule=_ru) == ret

    mock = MagicMock(return_value="table ip filter")
    ret = {
        "result": False,
        "comment": "Chain input in table filter in family ipv4 does not exist",
    }
    with patch.dict(nftables.__salt__, {"cmd.run": mock}):
        assert nftables.check(chain="input", rule=_ru) == ret

    mock = MagicMock(return_value="table ip filter chain input {{")
    ret = {
        "result": False,
        "comment": (
            "Rule tcp dport 22 log accept in chain input in table filter in family"
            " ipv4 does not exist"
        ),
    }
    with patch.dict(nftables.__salt__, {"cmd.run": mock}):
        assert nftables.check(chain="input", rule=_ru) == ret

    r_val = "table ip filter chain input {{ input tcp dport 22 log accept #"
    mock = MagicMock(return_value=r_val)
    ret = {
        "result": True,
        "comment": (
            "Rule tcp dport 22 log accept in chain input in table filter in family"
            " ipv4 exists"
        ),
    }
    with patch.dict(nftables.__salt__, {"cmd.run": mock}):
        assert nftables.check(chain="input", rule=_ru) == ret


# 'check_chain' function tests: 1


def test_check_chain():
    """
    Test if it check for the existence of a chain in the table
    """
    assert nftables.check_chain() == {
        "result": False,
        "comment": "Chain needs to be specified",
    }

    mock = MagicMock(return_value="")
    ret = {
        "comment": "Chain input in table filter in family ipv4 does not exist",
        "result": False,
    }
    with patch.dict(nftables.__salt__, {"cmd.run": mock}):
        assert nftables.check_chain(chain="input") == ret

    mock = MagicMock(return_value="chain input {{")
    ret = {
        "comment": "Chain input in table filter in family ipv4 exists",
        "result": True,
    }
    with patch.dict(nftables.__salt__, {"cmd.run": mock}):
        assert nftables.check_chain(chain="input") == ret


# 'check_table' function tests: 1


def test_check_table():
    """
    Test if it check for the existence of a table
    """
    assert nftables.check_table() == {
        "result": False,
        "comment": "Table needs to be specified",
    }

    mock = MagicMock(return_value="")
    ret = {"comment": "Table nat in family ipv4 does not exist", "result": False}
    with patch.dict(nftables.__salt__, {"cmd.run": mock}):
        assert nftables.check_table(table="nat") == ret

    mock = MagicMock(return_value="table ip nat")
    ret = {"comment": "Table nat in family ipv4 exists", "result": True}
    with patch.dict(nftables.__salt__, {"cmd.run": mock}):
        assert nftables.check_table(table="nat") == ret


# 'new_table' function tests: 1


def test_new_table():
    """
    Test if it create new custom table.
    """
    assert nftables.new_table(table=None) == {
        "result": False,
        "comment": "Table needs to be specified",
    }

    mock = MagicMock(return_value="")
    ret = {"comment": "Table nat in family ipv4 created", "result": True}
    with patch.dict(nftables.__salt__, {"cmd.run": mock}):
        assert nftables.new_table(table="nat") == ret

    mock = MagicMock(return_value="table ip nat")
    ret = {"comment": "Table nat in family ipv4 exists", "result": True}
    with patch.dict(nftables.__salt__, {"cmd.run": mock}):
        assert nftables.new_table(table="nat") == ret


# 'delete_table' function tests: 1


def test_delete_table():
    """
    Test if it delete custom table.
    """
    assert nftables.delete_table(table=None) == {
        "result": False,
        "comment": "Table needs to be specified",
    }

    mock_ret = {
        "result": False,
        "comment": "Table nat in family ipv4 does not exist",
    }
    with patch("salt.modules.nftables.check_table", MagicMock(return_value=mock_ret)):
        ret = nftables.delete_table(table="nat")
        assert ret == {
            "result": False,
            "comment": "Table nat in family ipv4 does not exist",
        }

    mock = MagicMock(return_value="table ip nat")
    with patch.dict(nftables.__salt__, {"cmd.run": mock}), patch(
        "salt.modules.nftables.check_table",
        MagicMock(return_value={"result": True, "comment": ""}),
    ):
        assert nftables.delete_table(table="nat") == {
            "comment": "Table nat in family ipv4 could not be deleted",
            "result": False,
        }

    mock = MagicMock(return_value="")
    with patch.dict(nftables.__salt__, {"cmd.run": mock}), patch(
        "salt.modules.nftables.check_table",
        MagicMock(return_value={"result": True, "comment": ""}),
    ):
        assert nftables.delete_table(table="nat") == {
            "comment": "Table nat in family ipv4 deleted",
            "result": True,
        }


# 'new_chain' function tests: 2


def test_new_chain():
    """
    Test if it create new chain to the specified table.
    """
    assert nftables.new_chain() == {
        "result": False,
        "comment": "Chain needs to be specified",
    }

    ret = {"result": False, "comment": "Table filter in family ipv4 does not exist"}
    mock = MagicMock(return_value="")
    with patch.dict(nftables.__salt__, {"cmd.run": mock}):
        assert nftables.new_chain(chain="input") == ret

    ret = {
        "result": False,
        "comment": "Chain input in table filter in family ipv4 already exists",
    }
    mock = MagicMock(return_value="table ip filter chain input {{")
    with patch.dict(nftables.__salt__, {"cmd.run": mock}):
        assert nftables.new_chain(chain="input") == ret


def test_new_chain_variable():
    """
    Test if it create new chain to the specified table.
    """
    mock = MagicMock(return_value="")
    with patch.dict(nftables.__salt__, {"cmd.run": mock}), patch(
        "salt.modules.nftables.check_chain",
        MagicMock(return_value={"result": False, "comment": ""}),
    ), patch(
        "salt.modules.nftables.check_table",
        MagicMock(return_value={"result": True, "comment": ""}),
    ):
        assert nftables.new_chain(chain="input", table_type="filter") == {
            "result": False,
            "comment": "Table_type, hook, and priority required.",
        }

        assert nftables.new_chain(
            chain="input", table_type="filter", hook="input", priority=0
        )


# 'delete_chain' function tests: 1


def test_delete_chain():
    """
    Test if it delete the chain from the specified table.
    """
    assert nftables.delete_chain() == {
        "result": False,
        "comment": "Chain needs to be specified",
    }

    ret = {"result": False, "comment": "Table filter in family ipv4 does not exist"}
    mock = MagicMock(return_value="")
    with patch.dict(nftables.__salt__, {"cmd.run": mock}):
        assert nftables.delete_chain(chain="input") == ret

    ret = {
        "result": False,
        "comment": ("Chain input in table filter in family ipv4 could not be deleted"),
    }
    mock = MagicMock(return_value="table ip filter")
    with patch.dict(nftables.__salt__, {"cmd.run": mock}), patch(
        "salt.modules.nftables.check_table",
        MagicMock(return_value={"result": True, "comment": ""}),
    ), patch(
        "salt.modules.nftables.check_chain",
        MagicMock(return_value={"result": True, "comment": ""}),
    ):
        assert nftables.delete_chain(chain="input") == ret

    ret = {
        "result": True,
        "comment": "Chain input in table filter in family ipv4 deleted",
    }
    mock = MagicMock(return_value="")
    with patch.dict(nftables.__salt__, {"cmd.run": mock}), patch(
        "salt.modules.nftables.check_table",
        MagicMock(return_value={"result": True, "comment": ""}),
    ), patch(
        "salt.modules.nftables.check_chain",
        MagicMock(return_value={"result": True, "comment": ""}),
    ):
        assert nftables.delete_chain(chain="input") == ret


def test_delete_chain_variables():
    """
    Test if it delete the chain from the specified table.
    """
    mock = MagicMock(return_value="")
    with patch.dict(nftables.__salt__, {"cmd.run": mock}), patch(
        "salt.modules.nftables.check_chain",
        MagicMock(return_value={"result": True, "comment": ""}),
    ), patch(
        "salt.modules.nftables.check_table",
        MagicMock(return_value={"result": True, "comment": ""}),
    ):
        _expected = {
            "comment": "Chain input in table filter in family ipv4 deleted",
            "result": True,
        }
        assert nftables.delete_chain(chain="input") == _expected


# 'append' function tests: 2


def test_append():
    """
    Test if it append a rule to the specified table & chain.
    """
    assert nftables.append() == {
        "result": False,
        "comment": "Chain needs to be specified",
    }

    assert nftables.append(chain="input") == {
        "result": False,
        "comment": "Rule needs to be specified",
    }

    _ru = "input tcp dport 22 log accept"
    ret = {"comment": "Table filter in family ipv4 does not exist", "result": False}
    mock = MagicMock(return_value="")
    with patch.dict(nftables.__salt__, {"cmd.run": mock}):
        assert nftables.append(chain="input", rule=_ru) == ret

    ret = {
        "comment": "Chain input in table filter in family ipv4 does not exist",
        "result": False,
    }
    mock = MagicMock(return_value="table ip filter")
    with patch.dict(nftables.__salt__, {"cmd.run": mock}):
        assert nftables.append(chain="input", rule=_ru) == ret

    r_val = "table ip filter chain input {{ input tcp dport 22 log accept #"
    mock = MagicMock(return_value=r_val)
    _expected = {
        "comment": (
            "Rule input tcp dport 22 log accept chain input in table filter in"
            " family ipv4 already exists"
        ),
        "result": False,
    }
    with patch.dict(nftables.__salt__, {"cmd.run": mock}):
        assert nftables.append(chain="input", rule=_ru) == _expected


def test_append_rule():
    """
    Test if it append a rule to the specified table & chain.
    """
    _ru = "input tcp dport 22 log accept"
    mock = MagicMock(side_effect=["1", ""])
    with patch.dict(nftables.__salt__, {"cmd.run": mock}), patch(
        "salt.modules.nftables.check",
        MagicMock(return_value={"result": False, "comment": ""}),
    ), patch(
        "salt.modules.nftables.check_chain",
        MagicMock(return_value={"result": True, "comment": ""}),
    ), patch(
        "salt.modules.nftables.check_table",
        MagicMock(return_value={"result": True, "comment": ""}),
    ):
        _expected = {
            "comment": (
                'Failed to add rule "{}" chain input in table filter in family'
                " ipv4.".format(_ru)
            ),
            "result": False,
        }
        assert nftables.append(chain="input", rule=_ru) == _expected
        _expected = {
            "comment": (
                'Added rule "{}" chain input in table filter in family ipv4.'.format(
                    _ru
                )
            ),
            "result": True,
        }
        assert nftables.append(chain="input", rule=_ru) == _expected


# 'insert' function tests: 2


def test_insert():
    """
    Test if it insert a rule into the specified table & chain,
    at the specified position.
    """
    assert nftables.insert() == {
        "result": False,
        "comment": "Chain needs to be specified",
    }

    assert nftables.insert(chain="input") == {
        "result": False,
        "comment": "Rule needs to be specified",
    }

    _ru = "input tcp dport 22 log accept"
    ret = {"result": False, "comment": "Table filter in family ipv4 does not exist"}
    mock = MagicMock(return_value="")
    with patch.dict(nftables.__salt__, {"cmd.run": mock}):
        assert nftables.insert(chain="input", rule=_ru) == ret

    ret = {
        "result": False,
        "comment": "Chain input in table filter in family ipv4 does not exist",
    }
    mock = MagicMock(return_value="table ip filter")
    with patch.dict(nftables.__salt__, {"cmd.run": mock}):
        assert nftables.insert(chain="input", rule=_ru) == ret

    r_val = "table ip filter chain input {{ input tcp dport 22 log accept #"
    mock = MagicMock(return_value=r_val)
    with patch.dict(nftables.__salt__, {"cmd.run": mock}):
        res = nftables.insert(chain="input", rule=_ru)
        import logging

        log = logging.getLogger(__name__)
        log.debug("=== res %s ===", res)
        assert nftables.insert(chain="input", rule=_ru)


def test_insert_rule():
    """
    Test if it insert a rule into the specified table & chain,
    at the specified position.
    """
    _ru = "input tcp dport 22 log accept"
    mock = MagicMock(side_effect=["1", ""])
    with patch.dict(nftables.__salt__, {"cmd.run": mock}), patch(
        "salt.modules.nftables.check",
        MagicMock(return_value={"result": False, "comment": ""}),
    ), patch(
        "salt.modules.nftables.check_chain",
        MagicMock(return_value={"result": True, "comment": ""}),
    ), patch(
        "salt.modules.nftables.check_table",
        MagicMock(return_value={"result": True, "comment": ""}),
    ):
        _expected = {
            "result": False,
            "comment": (
                'Failed to add rule "{}" chain input in table filter in family'
                " ipv4.".format(_ru)
            ),
        }
        assert nftables.insert(chain="input", rule=_ru) == _expected
        _expected = {
            "result": True,
            "comment": (
                'Added rule "{}" chain input in table filter in family ipv4.'.format(
                    _ru
                )
            ),
        }
        assert nftables.insert(chain="input", rule=_ru) == _expected


# 'delete' function tests: 2


def test_delete():
    """
    Test if it delete a rule from the specified table & chain,
    specifying either the rule in its entirety, or
    the rule's position in the chain.
    """
    _ru = "input tcp dport 22 log accept"
    ret = {
        "result": False,
        "comment": "Only specify a position or a rule, not both",
    }
    assert nftables.delete(table="filter", chain="input", position="3", rule=_ru) == ret

    ret = {"result": False, "comment": "Table filter in family ipv4 does not exist"}
    mock = MagicMock(return_value="")
    with patch.dict(nftables.__salt__, {"cmd.run": mock}):
        assert nftables.delete(table="filter", chain="input", rule=_ru) == ret

    ret = {
        "result": False,
        "comment": "Chain input in table filter in family ipv4 does not exist",
    }
    mock = MagicMock(return_value="table ip filter")
    with patch.dict(nftables.__salt__, {"cmd.run": mock}):
        assert nftables.delete(table="filter", chain="input", rule=_ru) == ret

    mock = MagicMock(return_value="table ip filter chain input {{")
    with patch.dict(nftables.__salt__, {"cmd.run": mock}):
        assert nftables.delete(table="filter", chain="input", rule=_ru)


def test_delete_rule():
    """
    Test if it delete a rule from the specified table & chain,
    specifying either the rule in its entirety, or
    the rule's position in the chain.
    """
    mock = MagicMock(side_effect=["1", ""])
    with patch.dict(nftables.__salt__, {"cmd.run": mock}), patch(
        "salt.modules.nftables.check",
        MagicMock(return_value={"result": True, "comment": ""}),
    ), patch(
        "salt.modules.nftables.check_chain",
        MagicMock(return_value={"result": True, "comment": ""}),
    ), patch(
        "salt.modules.nftables.check_table",
        MagicMock(return_value={"result": True, "comment": ""}),
    ):
        _expected = {
            "result": False,
            "comment": (
                'Failed to delete rule "None" in chain input  table filter in'
                " family ipv4"
            ),
        }
        assert nftables.delete(table="filter", chain="input", position="3") == _expected
        _expected = {
            "result": True,
            "comment": (
                'Deleted rule "None" in chain input in table filter in family ipv4.'
            ),
        }
        assert nftables.delete(table="filter", chain="input", position="3") == _expected


# 'flush' function tests: 2


def test_flush():
    """
    Test if it flush the chain in the specified table, flush all chains
    in the specified table if chain is not specified.
    """
    ret = {"result": False, "comment": "Table filter in family ipv4 does not exist"}
    mock = MagicMock(return_value="")
    with patch.dict(nftables.__salt__, {"cmd.run": mock}):
        assert nftables.flush(table="filter", chain="input") == ret

    ret = {
        "result": False,
        "comment": "Chain input in table filter in family ipv4 does not exist",
    }
    mock = MagicMock(return_value="table ip filter")
    with patch.dict(nftables.__salt__, {"cmd.run": mock}):
        assert nftables.flush(table="filter", chain="input") == ret


def test_flush_chain():
    """
    Test if it flush the chain in the specified table, flush all chains
    in the specified table if chain is not specified.
    """
    mock = MagicMock(side_effect=["1", ""])
    with patch.dict(nftables.__salt__, {"cmd.run": mock}), patch(
        "salt.modules.nftables.check_chain",
        MagicMock(return_value={"result": True, "comment": ""}),
    ), patch(
        "salt.modules.nftables.check_table",
        MagicMock(return_value={"result": True, "comment": ""}),
    ):
        _expected = {
            "result": False,
            "comment": (
                "Failed to flush rules from chain input in table filter in family"
                " ipv4."
            ),
        }
        assert nftables.flush(table="filter", chain="input") == _expected
        _expected = {
            "result": True,
            "comment": (
                "Flushed rules from chain input in table filter in family ipv4."
            ),
        }
        assert nftables.flush(table="filter", chain="input") == _expected


# 'get_policy' function tests: 1


def test_get_policy():
    """
    Test the current policy for the specified table/chain
    """
    list_rules_return = """
    {
      "nftables": [
        {
          "table": {
            "family": "ip",
            "name": "filter",
            "handle": 47
          }
        },
        {
          "chain": {
            "family": "ip",
            "table": "filter",
            "name": "input",
            "handle": 1,
            "type": "filter",
            "hook": "input",
            "prio": 0,
            "policy": "accept"
          }
        },
        {
          "chain": {
            "family": "ip",
            "table": "filter",
            "name": "forward",
            "handle": 2,
            "type": "filter",
            "hook": "forward",
            "prio": 0,
            "policy": "accept"
          }
        },
        {
          "chain": {
            "family": "ip",
            "table": "filter",
            "name": "output",
            "handle": 3,
            "type": "filter",
            "hook": "output",
            "prio": 0,
            "policy": "accept"
          }
        }
      ]
    }
    """
    expected = json.loads(list_rules_return)

    assert (
        nftables.get_policy(table="filter", chain=None, family="ipv4")
        == "Error: Chain needs to be specified"
    )

    with patch.object(nftables, "get_rules_json", MagicMock(return_value=expected)):
        assert (
            nftables.get_policy(table="filter", chain="input", family="ipv4")
            == "accept"
        )

    with patch.object(nftables, "get_rules_json", MagicMock(return_value=expected)):
        assert (
            nftables.get_policy(table="filter", chain="missing", family="ipv4") is None
        )


# 'set_policy' function tests: 1


def test_set_policy():
    """
    Test set the current policy for the specified table/chain
    """
    list_rules_return = """
    {
      "nftables": [
        {
          "table": {
            "family": "ip",
            "name": "filter",
            "handle": 47
          }
        },
        {
          "chain": {
            "family": "ip",
            "table": "filter",
            "name": "input",
            "handle": 1,
            "type": "filter",
            "hook": "input",
            "prio": 0,
            "policy": "accept"
          }
        },
        {
          "chain": {
            "family": "ip",
            "table": "filter",
            "name": "forward",
            "handle": 2,
            "type": "filter",
            "hook": "forward",
            "prio": 0,
            "policy": "accept"
          }
        },
        {
          "chain": {
            "family": "ip",
            "table": "filter",
            "name": "output",
            "handle": 3,
            "type": "filter",
            "hook": "output",
            "prio": 0,
            "policy": "accept"
          }
        }
      ]
    }
    """
    expected = json.loads(list_rules_return)["nftables"]

    assert (
        nftables.set_policy(table="filter", chain=None, policy=None, family="ipv4")
        == "Error: Chain needs to be specified"
    )

    assert (
        nftables.set_policy(table="filter", chain="input", policy=None, family="ipv4")
        == "Error: Policy needs to be specified"
    )

    mock = MagicMock(return_value={"retcode": 0})
    with patch.object(nftables, "get_rules_json", MagicMock(return_value=expected)):
        with patch.dict(nftables.__salt__, {"cmd.run_all": mock}):
            assert nftables.set_policy(
                table="filter", chain="input", policy="accept", family="ipv4"
            )



"""
Management of OpenStack Keystone Groups
=======================================

.. versionadded:: 2018.3.0

:depends: shade
:configuration: see :py:mod:`salt.modules.keystoneng` for setup instructions

Example States

.. code-block:: yaml

    create group:
      keystone_group.present:
        - name: group1

    delete group:
      keystone_group.absent:
        - name: group1

    create group with optional params:
      keystone_group.present:
        - name: group1
        - domain: domain1
        - description: 'my group'
"""


__virtualname__ = "keystone_group"


def __virtual__():
    if "keystoneng.group_get" in __salt__:
        return __virtualname__
    return (
        False,
        "The keystoneng execution module failed to load: shade python module is not"
        " available",
    )


def _common(kwargs):
    """
    Returns: None if group wasn't found, otherwise a group object
    """
    search_kwargs = {"name": kwargs["name"]}
    if "domain" in kwargs:
        domain = __salt__["keystoneng.get_entity"]("domain", name=kwargs.pop("domain"))
        domain_id = domain.id if hasattr(domain, "id") else domain
        search_kwargs["filters"] = {"domain_id": domain_id}
        kwargs["domain"] = domain

    return __salt__["keystoneng.group_get"](**search_kwargs)


def present(name, auth=None, **kwargs):
    """
    Ensure an group exists and is up-to-date

    name
        Name of the group

    domain
        The name or id of the domain

    description
        An arbitrary description of the group
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    __salt__["keystoneng.setup_cloud"](auth)

    kwargs = __utils__["args.clean_kwargs"](**kwargs)

    kwargs["name"] = name
    group = _common(kwargs)

    if group is None:
        if __opts__["test"] is True:
            ret["result"] = None
            ret["changes"] = kwargs
            ret["comment"] = "Group will be created."
            return ret

        group = __salt__["keystoneng.group_create"](**kwargs)
        ret["changes"] = group
        ret["comment"] = "Created group"
        return ret

    changes = __salt__["keystoneng.compare_changes"](group, **kwargs)
    if changes:
        if __opts__["test"] is True:
            ret["result"] = None
            ret["changes"] = changes
            ret["comment"] = "Group will be updated."
            return ret

        __salt__["keystoneng.group_update"](**kwargs)
        ret["changes"].update(changes)
        ret["comment"] = "Updated group"

    return ret


def absent(name, auth=None, **kwargs):
    """
    Ensure group does not exist

    name
        Name of the group

    domain
        The name or id of the domain
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    kwargs = __utils__["args.clean_kwargs"](**kwargs)

    __salt__["keystoneng.setup_cloud"](auth)

    kwargs["name"] = name
    group = _common(kwargs)

    if group:
        if __opts__["test"] is True:
            ret["result"] = None
            ret["changes"] = {"id": group.id}
            ret["comment"] = "Group will be deleted."
            return ret

        __salt__["keystoneng.group_delete"](name=group)
        ret["changes"]["id"] = group.id
        ret["comment"] = "Deleted group"

    return ret



"""
Configure a Pacemaker/Corosync cluster with PCS
===============================================

Configure Pacemaker/Cororsync clusters with the
Pacemaker/Cororsync conifguration system (PCS)

:depends: pcs

.. versionadded:: 2016.3.0
"""


import logging

import salt.utils.path

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load if pcs package is installed
    """
    if salt.utils.path.which("pcs"):
        return "pcs"
    return (False, "Missing dependency: pcs")


def __use_new_commands():
    """
    The command line arguments of pcs changed after version 0.10
    This will return True if the new arguments are needed and
    false if the old ones are needed
    """
    pcs_version = __salt__["pkg.version"]("pcs")
    log.debug("PCS package version %s", pcs_version)

    if __salt__["pkg.version_cmp"](pcs_version, "0.10") == 1:
        log.debug("New version, new command")
        return True
    else:
        log.debug("Old Version")
        return False


def item_show(
    item, item_id=None, item_type=None, show="show", extra_args=None, cibfile=None
):
    """
    Show an item via pcs command
    (mainly for use with the pcs state module)

    item
        config, property, resource, constraint etc.
    item_id
        id of the item
    item_type
        item type
    show
        show command (probably None, default: show or status for newer implementation)
    extra_args
        additional options for the pcs command
    cibfile
        use cibfile instead of the live CIB
    """

    new_commands = __use_new_commands()

    cmd = ["pcs"]

    if isinstance(cibfile, str):
        cmd += ["-f", cibfile]

    if isinstance(item, str):
        cmd += [item]
    elif isinstance(item, (list, tuple)):
        cmd += item

    # constraint command follows a different order
    if item in ["constraint"]:
        cmd += [item_type]

    # New implementions use config instead of show. This resolves that issue.
    if new_commands and (
        item != "config" and item != "constraint" and item != "property"
    ):
        if show == "show":
            show = "config"
        elif isinstance(show, (list, tuple)):
            for index, value in enumerate(show):
                if show[index] == "show":
                    show[index] = "config"

    if isinstance(show, str):
        cmd += [show]
    elif isinstance(show, (list, tuple)):
        cmd += show

    if isinstance(item_id, str):
        cmd += [item_id]

    if isinstance(extra_args, (list, tuple)):
        cmd += extra_args

    # constraint command only shows id, when using '--full'-parameter
    if item in ["constraint"]:
        if not isinstance(extra_args, (list, tuple)) or "--full" not in extra_args:
            cmd += ["--full"]
    log.debug("Running item show %s", cmd)
    return __salt__["cmd.run_all"](cmd, output_loglevel="trace", python_shell=False)


def item_create(
    item, item_id, item_type, create="create", extra_args=None, cibfile=None
):
    """
    Create an item via pcs command
    (mainly for use with the pcs state module)

    item
        config, property, resource, constraint etc.
    item_id
        id of the item
    item_type
        item type
    create
        create command (create or set f.e., default: create)
    extra_args
        additional options for the pcs command
    cibfile
        use cibfile instead of the live CIB
    """
    cmd = ["pcs"]
    if isinstance(cibfile, str):
        cmd += ["-f", cibfile]

    if isinstance(item, str):
        cmd += [item]
    elif isinstance(item, (list, tuple)):
        cmd += item

    # constraint command follows a different order
    if item in ["constraint"]:
        if isinstance(item_type, str):
            cmd += [item_type]

    if isinstance(create, str):
        cmd += [create]
    elif isinstance(create, (list, tuple)):
        cmd += create

    # constraint command needs item_id in format 'id=<id' after all params
    # constraint command follows a different order
    if item not in ["constraint"]:
        cmd += [item_id]
        if isinstance(item_type, str):
            cmd += [item_type]

    if isinstance(extra_args, (list, tuple)):
        # constraint command needs item_id in format 'id=<id' after all params
        if item in ["constraint"]:
            extra_args = extra_args + ["id={}".format(item_id)]
        cmd += extra_args

    return __salt__["cmd.run_all"](cmd, output_loglevel="trace", python_shell=False)


def auth(nodes, pcsuser="hacluster", pcspasswd="hacluster", extra_args=None):
    """
    Authorize nodes to the cluster

    nodes
        a list of nodes which should be authorized to the cluster
    pcsuser
        user for communitcation with PCS (default: hacluster)
    pcspasswd
        password for pcsuser (default: hacluster)
    extra_args
        list of extra option for the \'pcs cluster auth\' command. The newer cluster host command has no extra args and so will ignore it.

    CLI Example:

    .. code-block:: bash

        salt '*' pcs.auth nodes='[ node1.example.org, node2.example.org ]' pcsuser=hacluster pcspasswd=hoonetorg extra_args=[ '--force' ]
    """
    if __use_new_commands():
        cmd = ["pcs", "host", "auth"]
    else:
        cmd = ["pcs", "cluster", "auth"]

    cmd.extend(["-u", pcsuser, "-p", pcspasswd])

    if not __use_new_commands() and isinstance(extra_args, (list, tuple)):
        cmd += extra_args

    cmd += nodes

    return __salt__["cmd.run_all"](cmd, output_loglevel="trace", python_shell=False)


def is_auth(nodes, pcsuser="hacluster", pcspasswd="hacluster"):
    """
    Check if nodes are already authorized

    nodes
        a list of nodes to be checked for authorization to the cluster
    pcsuser
        user for communitcation with PCS (default: hacluster)
    pcspasswd
        password for pcsuser (default: hacluster)

    CLI Example:

    .. code-block:: bash

        salt '*' pcs.is_auth nodes='[node1.example.org, node2.example.org]' pcsuser=hacluster pcspasswd=hoonetorg
    """
    if __use_new_commands():
        cmd = ["pcs", "host", "auth", "-u", pcsuser, "-p", pcspasswd]
    else:
        cmd = ["pcs", "cluster", "auth"]

    cmd += nodes

    return __salt__["cmd.run_all"](
        cmd, stdin="\n\n", output_loglevel="trace", python_shell=False
    )


def cluster_setup(nodes, pcsclustername="pcscluster", extra_args=None):
    """
    Setup pacemaker cluster via pcs command

    nodes
        a list of nodes which should be set up
    pcsclustername
        Name of the Pacemaker cluster (default: pcscluster)
    extra_args
        list of extra option for the \'pcs cluster setup\' command

    CLI Example:

    .. code-block:: bash

        salt '*' pcs.cluster_setup nodes='[ node1.example.org, node2.example.org ]' pcsclustername=pcscluster
    """
    cmd = ["pcs", "cluster", "setup"]

    if __use_new_commands():
        cmd += [pcsclustername]
    else:
        cmd += ["--name", pcsclustername]

    cmd += nodes
    if isinstance(extra_args, (list, tuple)):
        cmd += extra_args

    log.debug("Running cluster setup: %s", cmd)

    return __salt__["cmd.run_all"](cmd, output_loglevel="trace", python_shell=False)


def cluster_destroy(extra_args=None):
    """
    Destroy corosync cluster using the pcs command

    extra_args
        list of extra option for the \'pcs cluster destroy\' command (only really --all)

    CLI Example:

    .. code-block:: bash

        salt '*' pcs.cluster_destroy extra_args=--all
    """
    cmd = ["pcs", "cluster", "destroy"]

    if isinstance(extra_args, (list, tuple)):
        cmd += extra_args

    log.debug("Running cluster destroy: %s", cmd)

    return __salt__["cmd.run_all"](cmd, output_loglevel="trace", python_shell=False)


def cluster_node_add(node, extra_args=None):
    """
    Add a node to the pacemaker cluster via pcs command

    node
        node that should be added
    extra_args
        list of extra option for the \'pcs cluster node add\' command

    CLI Example:

    .. code-block:: bash

        salt '*' pcs.cluster_node_add node=node2.example.org
    """
    cmd = ["pcs", "cluster", "node", "add"]

    cmd += [node]
    if isinstance(extra_args, (list, tuple)):
        cmd += extra_args

    return __salt__["cmd.run_all"](cmd, output_loglevel="trace", python_shell=False)


def cib_create(cibfile, scope="configuration", extra_args=None):
    """
    Create a CIB-file from the current CIB of the cluster

    cibfile
        name/path of the file containing the CIB
    scope
        specific section of the CIB (default: configuration)
    extra_args
        additional options for creating the CIB-file

    CLI Example:

    .. code-block:: bash

        salt '*' pcs.cib_create cibfile='/tmp/VIP_apache_1.cib' scope=False
    """
    cmd = ["pcs", "cluster", "cib", cibfile]
    if isinstance(scope, str):
        cmd += ["scope={}".format(scope)]
    if isinstance(extra_args, (list, tuple)):
        cmd += extra_args

    return __salt__["cmd.run_all"](cmd, output_loglevel="trace", python_shell=False)


def cib_push(cibfile, scope="configuration", extra_args=None):
    """
    Push a CIB-file as the new CIB to the cluster

    cibfile
        name/path of the file containing the CIB
    scope
        specific section of the CIB (default: configuration)
    extra_args
        additional options for creating the CIB-file

    CLI Example:

    .. code-block:: bash

        salt '*' pcs.cib_push cibfile='/tmp/VIP_apache_1.cib' scope=False
    """
    cmd = ["pcs", "cluster", "cib-push", cibfile]
    if isinstance(scope, str):
        cmd += ["scope={}".format(scope)]
    if isinstance(extra_args, (list, tuple)):
        cmd += extra_args

    return __salt__["cmd.run_all"](cmd, output_loglevel="trace", python_shell=False)


def config_show(cibfile=None):
    """
    Show config of cluster

    cibfile
        name/path of the file containing the CIB

    CLI Example:

    .. code-block:: bash

        salt '*' pcs.config_show cibfile='/tmp/cib_for_galera'
    """
    return item_show(item="config", item_id=None, extra_args=None, cibfile=cibfile)


def prop_show(prop, extra_args=None, cibfile=None):
    """
    Show the value of a cluster property

    prop
        name of the property
    extra_args
        additional options for the pcs property command
    cibfile
        use cibfile instead of the live CIB

    CLI Example:

    .. code-block:: bash

        salt '*' pcs.prop_show cibfile='/tmp/2_node_cluster.cib' prop='no-quorum-policy' cibfile='/tmp/2_node_cluster.cib'
    """
    return item_show(
        item="property", item_id=prop, extra_args=extra_args, cibfile=cibfile
    )


def prop_set(prop, value, extra_args=None, cibfile=None):
    """
    Set the value of a cluster property

    prop
        name of the property
    value
        value of the property prop
    extra_args
        additional options for the pcs property command
    cibfile
        use cibfile instead of the live CIB

    CLI Example:

    .. code-block:: bash

        salt '*' pcs.prop_set prop='no-quorum-policy' value='ignore' cibfile='/tmp/2_node_cluster.cib'
    """
    return item_create(
        item="property",
        item_id="{}={}".format(prop, value),
        item_type=None,
        create="set",
        extra_args=extra_args,
        cibfile=cibfile,
    )


def stonith_show(stonith_id, extra_args=None, cibfile=None):
    """
    Show the value of a cluster stonith

    stonith_id
        name for the stonith resource
    extra_args
        additional options for the pcs stonith command
    cibfile
        use cibfile instead of the live CIB

    CLI Example:

    .. code-block:: bash

        salt '*' pcs.stonith_show stonith_id='eps_fence' cibfile='/tmp/2_node_cluster.cib'
    """
    return item_show(
        item="stonith", item_id=stonith_id, extra_args=extra_args, cibfile=cibfile
    )


def stonith_create(
    stonith_id, stonith_device_type, stonith_device_options=None, cibfile=None
):
    """
    Create a stonith resource via pcs command

    stonith_id
        name for the stonith resource
    stonith_device_type
        name of the stonith agent fence_eps, fence_xvm f.e.
    stonith_device_options
        additional options for creating the stonith resource
    cibfile
        use cibfile instead of the live CIB for manipulation

    CLI Example:

    .. code-block:: bash

        salt '*' pcs.stonith_create stonith_id='eps_fence' stonith_device_type='fence_eps'
                                    stonith_device_options="['pcmk_host_map=node1.example.org:01;node2.example.org:02', 'ipaddr=myepsdevice.example.org', 'action=reboot', 'power_wait=5', 'verbose=1', 'debug=/var/log/pcsd/eps_fence.log', 'login=hidden', 'passwd=hoonetorg']" cibfile='/tmp/cib_for_stonith.cib'
    """
    return item_create(
        item="stonith",
        item_id=stonith_id,
        item_type=stonith_device_type,
        extra_args=stonith_device_options,
        cibfile=cibfile,
    )


def resource_show(resource_id, extra_args=None, cibfile=None):
    """
    Show a resource via pcs command

    resource_id
        name of the resource
    extra_args
        additional options for the pcs command
    cibfile
        use cibfile instead of the live CIB

    CLI Example:

    .. code-block:: bash

        salt '*' pcs.resource_show resource_id='galera' cibfile='/tmp/cib_for_galera.cib'
    """
    return item_show(
        item="resource", item_id=resource_id, extra_args=extra_args, cibfile=cibfile
    )


def resource_create(resource_id, resource_type, resource_options=None, cibfile=None):
    """
    Create a resource via pcs command

    resource_id
        name for the resource
    resource_type
        resource type (f.e. ocf:heartbeat:IPaddr2 or VirtualIP)
    resource_options
        additional options for creating the resource
    cibfile
        use cibfile instead of the live CIB for manipulation

    CLI Example:

    .. code-block:: bash

        salt '*' pcs.resource_create resource_id='galera' resource_type='ocf:heartbeat:galera' resource_options="['wsrep_cluster_address=gcomm://node1.example.org,node2.example.org,node3.example.org', '--master']" cibfile='/tmp/cib_for_galera.cib'
    """
    return item_create(
        item="resource",
        item_id=resource_id,
        item_type=resource_type,
        extra_args=resource_options,
        cibfile=cibfile,
    )



"""
Execution module for RESTCONF Proxy minions

:codeauthor: Jamie (Bear) Murphy <jamiemurphyit@gmail.com>
:maturity:   new
:platform:   any

"""


import logging

__proxyenabled__ = ["restconf"]
__virtualname__ = "restconf"

log = logging.getLogger(__file__)


def __virtual__():
    if __opts__.get("proxy", {}).get("proxytype") != __virtualname__:
        return False, "Proxytype does not match: {}".format(__virtualname__)
    return True


def info():
    """
    Returns the RESTCONF capabilities PATH

    CLI Example:

    .. code-block:: bash

        salt '*' restconf.info
    """
    return __proxy__["restconf.request"](
        "restconf/data/ietf-restconf-monitoring:restconf-state/capabilities/capability"
    )


def get_data(path):
    """
    Returns an object containing the content of the request path with a GET request.
    Data returned will contain a dict with at minimum a key of "status" containing the http status code
    Other keys that should be available error (if http error), body, dict (parsed json to dict)

    CLI Example:

    .. code-block:: bash

        salt '*' restconf.get_data restconf/yang-library-version
    """
    return __proxy__["restconf.request"](path)


def set_data(path, method, dict_payload):
    """
    Sends a post/patch/other type of rest method to a specified path with the specified method with specified payload

    CLI Example:

    .. code-block:: bash

        salt '*' restconf.set_data restconf/yang-library-version method=PATCH dict_payload=""
    """
    return __proxy__["restconf.request"](path, method, dict_payload)


def path_check(primary_path, init_path):
    """
    Used to check which path responds with a 200 status
    Returns an array of True/False and a dict with keys path + path_method + response data, used in states code.

    CLI Example:

    .. code-block:: bash

        salt '*' restconf.path_check restconf/yang-library-version/specifc_item restconf/yang-library-version
    """
    ret = {"result": False}

    log.debug("modules_restconf_path_check: about to attempt to get primary path")
    existing_raw = __salt__["restconf.get_data"](primary_path)

    if existing_raw["status"] == 200:
        log.debug("modules_restconf_path_check: found a valid path at primary_path")
        existing = existing_raw["dict"]
        ret["result"] = True
        ret["path_used"] = "primary"
        ret["request_path"] = primary_path
        ret["request_restponse"] = existing

    if not ret["result"]:
        if init_path is not None:
            existing_raw_init = __salt__["restconf.get_data"](init_path)
            if existing_raw_init["status"] in [200]:
                log.debug(
                    "modules_restconf_path_check: found a valid path at init_path"
                )
                existing = existing_raw_init["dict"]
                ret["result"] = True
                ret["path_used"] = "init"
                ret["request_path"] = init_path
                ret["request_restponse"] = existing

    if not ret["result"]:
        log.debug(
            "modules_restconf_path_check: restconf could not find a working path to get initial config"
        )
        ret["result"] = False

    return ret



"""
Return salt data to Nagios

The following fields can be set in the minion conf file::

    nagios.url (required)
    nagios.token (required)
    nagios.service (optional)
    nagios.check_type (optional)

Alternative configuration values can be used by prefacing the configuration.
Any values not found in the alternative configuration will be pulled from
the default location::

    nagios.url
    nagios.token
    nagios.service

Nagios settings may also be configured as::

    nagios:
        url: http://localhost/nrdp
        token: r4nd0mt0k3n
        service: service-check

    alternative.nagios:
        url: http://localhost/nrdp
        token: r4nd0mt0k3n
        service: another-service-check

  To use the Nagios returner, append '--return nagios' to the salt command. ex:

  .. code-block:: bash

    salt '*' test.ping --return nagios

  To use the alternative configuration, append '--return_config alternative' to the salt command. ex:

    salt '*' test.ping --return nagios --return_config alternative

To override individual configuration items, append --return_kwargs '{"key:": "value"}' to the salt command.

.. versionadded:: 2016.3.0

.. code-block:: bash

    salt '*' test.ping --return nagios --return_kwargs '{"service": "service-name"}'

"""

import html
import http.client
import logging

import salt.returners

log = logging.getLogger(__name__)

__virtualname__ = "nagios_nrdp"


def __virtual__():
    """
    Return virtualname
    """
    return "nagios.list_plugins" in __salt__


def _get_options(ret=None):
    """
    Get the requests options from salt.
    """
    attrs = {
        "url": "url",
        "token": "token",
        "service": "service",
        "checktype": "checktype",
    }

    _options = salt.returners.get_returner_options(
        __virtualname__, ret, attrs, __salt__=__salt__, __opts__=__opts__
    )

    log.debug("attrs %s", attrs)
    if "checktype" not in _options or _options["checktype"] == "":
        # default to passive check type
        _options["checktype"] = "1"

    if _options["checktype"] == "active":
        _options["checktype"] = "0"

    if _options["checktype"] == "passive":
        _options["checktype"] = "1"

    # checktype should be a string
    _options["checktype"] = str(_options["checktype"])

    return _options


def _prepare_xml(options=None, state=None):
    """
    Get the requests options from salt.
    """

    if state:
        _state = "0"
    else:
        _state = "2"

    xml = "<?xml version='1.0'?>\n<checkresults>\n"

    # No service defined then we set the status of the hostname
    if options.get("service"):
        xml += (
            "<checkresult type='service' checktype='" + str(options["checktype"]) + "'>"
        )
        xml += "<hostname>" + html.escape(options["hostname"]) + "</hostname>"
        xml += "<servicename>" + html.escape(options["service"]) + "</servicename>"
    else:
        xml += "<checkresult type='host' checktype='" + str(options["checktype"]) + "'>"
        xml += "<hostname>" + html.escape(options["hostname"]) + "</hostname>"

    xml += "<state>" + _state + "</state>"

    if options.get("output"):
        xml += "<output>" + html.escape(options["output"]) + "</output>"

    xml += "</checkresult>"

    xml += "\n</checkresults>"

    return xml


def _getText(nodelist):
    """
    Simple function to return value from XML
    """
    rc = []
    for node in nodelist:
        if node.nodeType == node.TEXT_NODE:
            rc.append(node.data)
    return "".join(rc)


def _post_data(options=None, xml=None):
    """
    Post data to Nagios NRDP
    """
    params = {"token": options["token"].strip(), "cmd": "submitcheck", "XMLDATA": xml}

    res = salt.utils.http.query(
        url=options["url"],
        method="POST",
        params=params,
        data="",
        decode=True,
        status=True,
        header_dict={},
        opts=__opts__,
    )

    if res.get("status", None) == http.client.OK:
        if res.get("dict", None) and isinstance(res["dict"], list):
            _content = res["dict"][0]
            if _content.get("status", None):
                return True
            else:
                return False
        else:
            log.error("No content returned from Nagios NRDP.")
            return False
    else:
        log.error("Error returned from Nagios NRDP. Status code: %s.", res.status_code)
        return False


def returner(ret):
    """
    Send a message to Nagios with the data
    """

    _options = _get_options(ret)
    log.debug("_options %s", _options)
    _options["hostname"] = ret.get("id")

    if "url" not in _options or _options["url"] == "":
        log.error("nagios_nrdp.url not defined in salt config")
        return

    if "token" not in _options or _options["token"] == "":
        log.error("nagios_nrdp.token not defined in salt config")
        return

    xml = _prepare_xml(options=_options, state=ret["return"])
    res = _post_data(options=_options, xml=xml)

    return res
